package net.sourceforge.amore.gui.Actions ;

import net.sourceforge.amore.gui.global.* ; // Resource, EditMethods
import net.sourceforge.amore.gui.AmoreFrame ;
import net.sourceforge.amore.gui.SessionTree.* ;
import net.sourceforge.amore.gui.SessionTree.GuiNode.* ;

import net.sourceforge.amore.gui.Dialogs.* ;

import net.sourceforge.amore.RegLangSet ;
import net.sourceforge.amore.Ldo.* ;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import java.io.File;
import javax.swing.filechooser.* ;
import net.sourceforge.amore.gui.util.ExampleFileFilter ;

public class FileAction {
    static AmoreFrame frame = AmoreFrame.getInstance();
    static SessionTree sessionTree = SessionTree.getInstance();

    static void message(String s) {
	AmoreFrame.setStatusText(s);
    }

    static void update() {
	AmoreFrame.update();
    }

    /*  zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz  FILE / NEW
     */
    
    static class FileNew extends AbstractAction {
	public void actionPerformed(ActionEvent e) {
	    // create 
	    message("Creating new LDO");
	    
	    RegLangSet newBundle = CreateModifyDialog.showDialog( (Component) null, null );
	    if( newBundle != null ) {
		sessionTree.addRegLangSet( newBundle );
		message(" Created new "+EditMethods.getNameTerse( newBundle.getInputClass() ));
		update();
	    } else {
		message("cancelled.");
	    }
	}
	public FileNew() {
	    super();
	    putValue(Action.NAME, "New ...");
	    setEnabled(true);
	}
    }
    static FileNew fileNew = new FileNew();

    /*  zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz  FILE / EDIT
     */
    
    static class FileEdit extends AbstractAction {
	public void actionPerformed(ActionEvent e) {
	    GuiNode n;
	    message("Editing");
	    // get selected node
	    n = sessionTree.getCurrent();
	    final AmoreFrame aFrame = AmoreFrame.getInstance();
	    
            RegLangSet oldRs = (RegLangSet) n.apply(new GuiNodeVisitor() {
		    public Object caseLwpNode(LwpNode l) {
			LangDefObj ldo = l.getLdo();
			// we cannot edit monoids.
			if (!(ldo instanceof Monoid))
			    return new RegLangSet( (LangDefObj) l.getLdo().clone() );
			message(" you cannot edit monoids");
			return null;
		    }
		    public Object caseRootNode( RootNode r) {
			message(" please select a language.");
			return null;
		    }
		    public Object caseRegLangSetNode(RegLangSetNode r) {
			return r.getRegLangSet();
		    }
		});
            
            if(oldRs != null) {
                RegLangSet newRs = CreateModifyDialog.showDialog( (Component) null, oldRs);
                if( newRs != null ) { //editing successful
                    sessionTree.removeCurrent();
                    sessionTree.addRegLangSet( newRs );
                    message("Created new language with "+EditMethods.getNameTerse( newRs.getInputClass() )+" as input object");
                    update();
                } else {
                    message("Editing cancelled.");
                }
            }
	}
	public FileEdit() {
	    super();
	    putValue(Action.NAME, "Edit ...");
	    setEnabled(true);
	}
    }
    static FileEdit fileEdit = new FileEdit();

    /*  zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz  FILE / LOAD
     */

    static class FileLoad extends AbstractAction {
	public void actionPerformed(ActionEvent e) {
	    message("Loading a regular language");
            JFileChooser chooser = new JFileChooser();
            chooser.setDialogTitle("choose file to load");
            chooser.setDialogType( JFileChooser.OPEN_DIALOG );
            chooser.setApproveButtonToolTipText("loads selected file");
            chooser.setFileSelectionMode( JFileChooser.FILES_ONLY );

            ExampleFileFilter filter = new ExampleFileFilter();
            filter.addExtension("amr");
            filter.setDescription("AMoRE files");
            chooser.setFileFilter(filter);

            int retVal = chooser.showOpenDialog( frame );
            if(retVal == JFileChooser.APPROVE_OPTION) {
                String fname = chooser.getSelectedFile().getName();
                String path = chooser.getSelectedFile().getPath();
                System.err.println("loading \""+path+"\"");
                // need fname without ".amr" extension
                fname= fname.substring(0,fname.lastIndexOf('.'));
                // need path without filename
                path =  path.substring(0,path.lastIndexOf('/'));
                System.err.println("calling load(\""+fname+"\",\""+path+"\"");
                RegLangSet r = RegLangSet.load(fname, path);
                if (r != null) {
                    sessionTree.addRegLangSet( r );
                    message("added "+r.getName()+" to session");
                    update();
                } else {
                    //error message
                }
            }
	}
	public FileLoad() {
	    super();
	    putValue(Action.NAME, "Load LDO");
	    setEnabled(true);
	}
    }
    static FileLoad fileLoad = new FileLoad();

    /*  zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz  FILE / SAVE
     */


    static class FileSave extends AbstractAction {
	public void actionPerformed(ActionEvent e) {
	    GuiNode n = sessionTree.getCurrent();
	    RegLangSet rs = (RegLangSet) n.apply( new GuiNodeVisitor() {  
                    public Object caseRootNode( RootNode r) {
			return null;
		    } 
		    public Object caseRegLangSetNode(RegLangSetNode r) {
			return r.getRegLangSet();
		    }
                    public Object caseLwpNode(LwpNode l) {
                        return l.getLdo().getRegSet();
                    }
                });
            
            if (rs==null) {
		message("you can only save whole languages"); 
		return;
            }
	    message("Saving a regular language");
            JFileChooser schooser = new JFileChooser();
            schooser.setDialogTitle("in which directory should I save \""+rs.getName()+"\"");
            schooser.setFileFilter( new FileFilter() {
                    public boolean accept(File f) { return true; }
                    public String getDescription() { 
                        return "choose a directory";
                    }
                } );
            schooser.setAcceptAllFileFilterUsed( false );
            schooser.setFileSelectionMode( JFileChooser.DIRECTORIES_ONLY );
            schooser.setApproveButtonToolTipText("saves "+ rs.getName() +" to the selected directory");
            int retuVal = schooser.showSaveDialog( frame );
            if(retuVal == JFileChooser.APPROVE_OPTION) {
                File sfile = schooser.getSelectedFile();
                String dir = schooser.getSelectedFile().toString();
                System.out.print(sfile);
                if (sfile.isDirectory()) {
                    message("You chose to save to this dir: " +
			    dir);
                    System.err.println("saving to \""+dir+"\""); 
                    
                    rs.save(dir);
                } else {
                    message(sfile.getName()+" is not a directory");
                }
            } else {
                message("cancelled.");
            }
	}
	public FileSave() {
	    super();
	    putValue(Action.NAME, "Save LDO");
	    setEnabled(true);
	}
    }
    static FileSave fileSave = new FileSave();
    
    
    /*  zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz  FILE / DELETE
     */
    static class FileDelete extends AbstractAction {
	public void actionPerformed(ActionEvent e) {
	    sessionTree.removeCurrent();
	    update();
	}
	public FileDelete() {
	    super();
	    putValue(Action.NAME, "Delete");
	    setEnabled(true);
	}
    }
    static FileDelete fileDelete = new FileDelete();
    
    /*  zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz  FILE / PRINT
     */
    static class FilePrint extends AbstractAction { 
	// N O T    I M P L E M E N T E D
	public void actionPerformed(ActionEvent e) {
	    System.err.println("N O T    I M P L E M E N T E D");
	    frame.show_NotImplemented_Box();
	}
	public FilePrint() {
	    super();
	    putValue(Action.NAME, "Print");
	    setEnabled(false);
	}
    }
    static FilePrint filePrint = new FilePrint();
    
    public static AbstractAction[] allActions = {
	fileNew,
	fileEdit,
	fileLoad,
	fileSave,
	fileDelete,
	filePrint
    };
    
}
