package net.sourceforge.amore.gui.Actions ;

import net.sourceforge.amore.gui.AmoreFrame ;
import net.sourceforge.amore.gui.SessionTree.SessionTree;
import net.sourceforge.amore.gui.SessionTree.GuiNode.*;

import net.sourceforge.amore.Ldo.*;
import net.sourceforge.amore.RegLangSet ;

import  net.sourceforge.amore.gui.Dialogs.ListDialog ;

import java.util.Vector ;
import java.util.Enumeration ;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**  this class holds action objects that represent binary transformations
 *  (e.g. L1 leftquot L2). is a singleton.
 */
public class TransfBinary extends Arrows {

    private static TransfBinary instance;

    private TransfBinary() {
}    

    public static TransfBinary getInstance() {
        if( instance == null ) {
            instance = new TransfBinary();
        }
        return instance;
    }
    public void toggleActions( AbstractAction[] ac, boolean b ) {
	for (int i=0; i<ac.length; i++) {
	    ac[i].setEnabled( b );
	}
    }

    private static Nfa askForNfa() {
        // show Dialog LdoChooser
        Vector choices = SessionTree.getInstance().getLanguages();// Vector of RegLangSet obj
        //System.out.println("askforNfa "+choices);
        // get rid of langs without nfa
        for(Enumeration e=choices.elements(); e.hasMoreElements(); ) {
            RegLangSet rs = (RegLangSet) e.nextElement();
            // this might be too restrictive but is backwards
            //   compatible with AMoRE native library
            if((rs.getNletters() != source.getNletters())
               ||(rs.getNfa() == null)
               ||(rs.getNfa().isEps()))
                choices.remove(rs);
        }
        //System.out.println("askforNfa "+choices);

        ListDialog.initialize(AmoreFrame.getInstance(), choices,
                              "choose language for L2",
                              "list of langs with nfa \n and same alphabet");
        RegLangSet selLang = (RegLangSet) ListDialog.showDialog(null,
                                                                    null);
        //System.out.println("selected "+selLang.getName()+ "as 2nd lan");
        return selLang.getNfa();
    }

    private static Dfa askForDfa() {
        Vector choices = SessionTree.getInstance().getLanguages();// Vector of RegLangSet obj
        // get rid of langs without dfa
        for(Enumeration e=choices.elements(); e.hasMoreElements(); ) {
            RegLangSet rs = (RegLangSet) e.nextElement();
            // this might be too restrictive but is backwards
            //   compatible with AMoRE native library
            if((rs.getNletters() != source.getNletters())
               ||(rs.getDfa() == null))
                choices.remove(rs);
        }
        //System.out.println("askforNfa "+choices);
        ListDialog.initialize(AmoreFrame.getInstance(), choices,
                              "choose language for L2",
                              "list of langs with dfa \n and same alphabet");
        RegLangSet selLang = (RegLangSet) ListDialog.showDialog(null,
                                                                null);
        return selLang.getDfa();
    }

    public static SessionTree sessionTree = SessionTree.getInstance();

    public static class UNION extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    Nfa n1 = (Nfa) source;
	    Nfa n2 = askForNfa();
            Nfa n = BinTransforms.unionNfa( n1, n2 );
            RegLangSet rset = new RegLangSet(n);
            rset.setName("union");
            rset.setDescr("L("+n1.getRegSet().getName()
                          +") U L("    +n2.getRegSet().getName()+")");

	    sessionTree.addRegLangSet( rset );
	}
	public  UNION() {
	    super();
	    putValue(Action.NAME, "Union of L(Nfa1), L(Nfa2)");
	    setEnabled(false);
	}
    }
    public static UNION union = new UNION();

    public static class CONCAT extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    Nfa n1 = (Nfa) source;
	    Nfa n2 = askForNfa();
            Nfa n = BinTransforms.concatNfa( n1, n2 );
            RegLangSet rset = new RegLangSet(n);
            rset.setName("concat");
            rset.setDescr("L("+n1.getRegSet().getName()
                          +") . L("    +n2.getRegSet().getName()+")");
	    sessionTree.addRegLangSet( rset );
	}
	public  CONCAT() {
	    super();
	    putValue(Action.NAME, "Concat of L(Nfa1), L(Nfa2)");
	    setEnabled(false);
	}
    }
    public static CONCAT concat = new CONCAT();


    public static class INSEC extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    Dfa d1 = (Dfa) source;
	    Dfa d2 = askForDfa();
            Dfa d = BinTransforms.insecDfa( d1, d2 );
            RegLangSet rset = new RegLangSet(d);
            rset.setName("insec");
	    sessionTree.addRegLangSet( rset );
	}
	public  INSEC() {
	    super();
	    putValue(Action.NAME, "Intersection L(Dfa1), L(Dfa2)");
	    setEnabled(false);
	}
    }
    public static INSEC insec = new INSEC();

    public static class MINUS extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    Dfa d1 = (Dfa) source;
	    Dfa d2 = askForDfa();
            Dfa d = BinTransforms.minusDfa( d1, d2 );
            RegLangSet rset = new RegLangSet(d);
            rset.setName("minus");
            rset.setDescr(" L("+d1.getRegSet().getName()
                          +") \\ L("    +d2.getRegSet().getName()+")");

	    sessionTree.addRegLangSet( rset );
	}
	public  MINUS() {
	    super();
	    putValue(Action.NAME, "L(Dfa1) -  L(Dfa2)");
	    setEnabled(false);
	}
    }
    public static MINUS minus = new MINUS();

    public static class SHUFFLE extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    Nfa n1 = (Nfa) source;
	    Nfa n2 = askForNfa();
            Nfa n = BinTransforms.shuffleNfa( n1, n2 );
            RegLangSet rset = new RegLangSet(n);
            rset.setName("shuffle");
            rset.setDescr("shuffle L("+n1.getRegSet().getName()
                          +") W L("    +n2.getRegSet().getName()+")");

	    sessionTree.addRegLangSet( rset );
	}
	public  SHUFFLE() {
	    super();
	    putValue(Action.NAME, "Shuffle of L(Nfa1), L(Nfa2)");
	    setEnabled(false);
	}
    }
    public static SHUFFLE shuffle = new SHUFFLE();

    public static class RIGHTQUOT extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    Dfa d1 = (Dfa) source;
	    Dfa d2 = askForDfa();
            Dfa d = BinTransforms.rightQuot( d1, d2 );
            RegLangSet rset = new RegLangSet(d);
            rset.setName("rightquot");
            rset.setDescr("rightquot L("+d1.getRegSet().getName()
                          +") RQ L("    +d2.getRegSet().getName()+")");
	    sessionTree.addRegLangSet( rset );
	}
	public  RIGHTQUOT() {
	    super();
	    putValue(Action.NAME, "right quotient L(Dfa1) RQ L(Dfa2)");
	    setEnabled(false);
	}
    }
    public static RIGHTQUOT rightquot = new RIGHTQUOT();

    public static class LEFTQUOT extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    Dfa d1 = (Dfa) source;
	    Dfa d2 = askForDfa();
            Dfa d = BinTransforms.leftQuot( d1, d2 );
            RegLangSet rset = new RegLangSet(d);
            rset.setName("leftquot");
            rset.setDescr("leftquot L("+d1.getRegSet().getName()
                          +") RQ L("    +d2.getRegSet().getName()+")");
	    sessionTree.addRegLangSet( rset );
	}
	public  LEFTQUOT() {
	    super();
	    putValue(Action.NAME, "left quotient  L(Dfa1) LQ L(Dfa2)");
	    setEnabled(false);
	}
    }
    public static LEFTQUOT leftquot = new LEFTQUOT();


    final static AbstractAction[] nfaActions = { 
        union, concat, shuffle };
    final static AbstractAction[] dfaActions = {
        insec, minus, rightquot, leftquot };


    final static AbstractAction allActions[] = {
        union, concat, shuffle,
        insec, minus, rightquot, leftquot
    };


    public AbstractAction[] getAllActions() {
        return allActions;
    }

    /**  toggles actions on or off, depending on type of source.
     *   If argument is null, all actions are toggled off.
     */
    protected void toggleActions() {
        if( source == null ) {

            toggleActions( allActions, false );

        } else {

            source.apply( new LangDefObjVisitor() {
                    public Object caseNfa(Nfa n) {
                        // handle epsNfa
                        if( n.isEps() ) {
                            toggleActions( allActions, false);
                        } else {
                            toggleActions( nfaActions, true);
                        }
                        toggleActions( dfaActions, false);
                        return null;
                    }
                    public Object caseDfa(Dfa d) {
                        toggleActions( nfaActions, false);
                        toggleActions( dfaActions, true);
                        return null;
                    }
                    public Object caseRegExp(RegExp r) {
                        toggleActions(allActions, false);
                        return null;
                    }
                    public Object caseMonoid(Monoid m) {
                        toggleActions(allActions, false);
                        return null;
                    }
                });
        } /* else */
    } /* toggleActions() */
    
}
