package net.sourceforge.amore.gui.Actions ;

import net.sourceforge.amore.Ldo.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/** this class holds action objects that represent unary transformations
 *  that preserve the regular language (e.g. NFA->DFA). is a singleton.
 */
public class TransfUnaryA extends Arrows {

    // singleton methods

    private static TransfUnaryA instance;

    private TransfUnaryA() {        
        toggleActions( allActions, false );
    }

    public static TransfUnaryA getInstance() {
        if( instance == null )
            instance = new TransfUnaryA();
        return instance;
    }

    //
    // ----------------------------------- RegExp actions
    //
    
    public static class GENREGEXP_TO_NFA extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    System.out.println("GENREGEXP_TO_NFA"); // -------- GENREGEXP -> NFA
	    RegExp r = (RegExp) source;
	    Nfa n = Transforms.genrex2nfa( r );
            r.getRegSet().addLdo( n ); 

	}
	public  GENREGEXP_TO_NFA() {
	    super();
	    putValue(Action.NAME, "GenRegExp to Nfa");
	    
	}
    }

    public static GENREGEXP_TO_NFA genRegExp2Nfa = new GENREGEXP_TO_NFA();
    
    public static class REGEXP_TO_NFA    extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    System.out.println("REGEXP_TO_NFA"); // ----------- REGEXP -> NFA
	    RegExp r = (RegExp) source;
	    Nfa n = Transforms.rex2nfa( r );
            r.getRegSet().addLdo( n );
	}
	public REGEXP_TO_NFA() {
	    super();
	    putValue(Action.NAME, "RegExp to Nfa");
	    
	}
    }
    public static REGEXP_TO_NFA regExp2Nfa = new REGEXP_TO_NFA();

    public final static AbstractAction regExpActions[] = { genRegExp2Nfa, 
							   regExp2Nfa };
    //
    // ---------------------------------------------------- NFA actions
    //

    public static class ENFA_TO_NFA extends AbstractAction {
	public void actionPerformed(ActionEvent e) {
	    System.out.println("ENFA_TO_NFA");// ------------- ENFA -> NFA
	    Nfa n = (Nfa) source;
	    Nfa n2 = Transforms.enfa2nfa( n );
            n.getRegSet().addLdo( n2 );
	}
	public ENFA_TO_NFA() {
	    super();
	    putValue(Action.NAME, "epsNfa to Nfa");
	    
	}
    }
    public static ENFA_TO_NFA eNfa2Nfa = new ENFA_TO_NFA();

    public static class NFA_TO_DFA       extends AbstractAction { 	
	public void actionPerformed(ActionEvent e) {
	    System.out.println("NFA_TO_DFA"); // ------------- NFA -> DFA
	    Nfa n = (Nfa) source;
	    Dfa d = Transforms.nfa2dfa( n );
            n.getRegSet().addLdo( d );
	}
	public NFA_TO_DFA() {
	    super();
	    putValue(Action.NAME, "Nfa to Dfa");
	    
	}
    }
    public static NFA_TO_DFA nfa2Dfa = new NFA_TO_DFA();

    public static class NFA_TO_REGEXP    extends AbstractAction {	
	public void actionPerformed(ActionEvent e) {
	    System.out.println("NFA_TO_REGEXP");// ---------- NFA -> REGEXP
	    Nfa n = (Nfa) source;
	    RegExp r = Transforms.nfa2rex( n );
            n.getRegSet().addLdo( r );
	}
	public NFA_TO_REGEXP() {
	    super();
	    putValue(Action.NAME, "Nfa to RegExp");
	    
	}
    }
    public static NFA_TO_REGEXP nfa2RegExp = new NFA_TO_REGEXP();
    
    public static class NFA_TO_REDNFA    extends AbstractAction {	
	public void actionPerformed(ActionEvent e) {
	    System.out.println("NFA_TO_REDNFA (needs mindfa)");
	    // ---------- NFA -> REDNFA
	    
	    Nfa n = (Nfa) source;
	    Dfa d = n.getRegSet().getMinDfa();
	    if( d != null ) {
		Nfa redn = Transforms.nfa2rednfa( n, d );
                n.getRegSet().addLdo( redn );
	    } else {
		System.out.println("mindfa not present");
	    }
		
	}
	public NFA_TO_REDNFA() {
	    super();
	    putValue(Action.NAME, "Nfa to RedNfa");
	    
	}
    }
    public static NFA_TO_REDNFA nfa2RedNfa = new NFA_TO_REDNFA();
       

    public static AbstractAction nfaActions[] = { eNfa2Nfa, 
						  nfa2RegExp, 
						  nfa2Dfa,
						  nfa2RedNfa };
  //
  // ---------------------------------------------------------- DFA Actions
  //

    public static class DFA_TO_NFA    extends AbstractAction { 	
	public void actionPerformed(ActionEvent e) {
	    System.out.println("DFA_TO_NFA");// ------------- DFA -> NFA
	    Dfa d = (Dfa) source;
	    Nfa n = Transforms.dfa2nfa( d );
            d.getRegSet().addLdo( n );
	}
	public DFA_TO_NFA() {
	    super();
	    putValue(Action.NAME, "Dfa to Nfa");
	    
	}
    }

    public static DFA_TO_NFA dfa2Nfa = new DFA_TO_NFA();

    public static class DFA_TO_MINDFA extends AbstractAction { 	
	public void actionPerformed(ActionEvent e) {
	    System.out.println("DFA_TO_MINDFA");// --------- DFA -> MINDFA
	    Dfa d = (Dfa) source;
	    Dfa md = Transforms.dfa2mindfa( d );	    
            d.getRegSet().addLdo( md );
	}
	public DFA_TO_MINDFA() {
	    super();
	    putValue(Action.NAME, "Dfa to MinDfa");
	    
	}
    }

    public static DFA_TO_MINDFA dfa2MinDfa = new DFA_TO_MINDFA();

    public static class DFA_TO_MONOID extends AbstractAction { 	
	public void actionPerformed(ActionEvent e) {
	    System.out.println("DFA_TO_MINDFA");// --------- DFA -> MINDFA
	    Dfa d    = (Dfa) source;
	    Monoid m = Transforms.dfa2mon( d );	    
            d.getRegSet().addLdo( m );
	}
	public DFA_TO_MONOID() {
	    super();
	    putValue(Action.NAME, "Dfa to Monoid");
	    
	}
    }

    public static DFA_TO_MONOID dfa2Monoid = new DFA_TO_MONOID();

    

    public static AbstractAction dfaActions[] = { dfa2Nfa, 
						  dfa2MinDfa,
						  dfa2Monoid };

    // ...

    final static AbstractAction[] allActions = {
        genRegExp2Nfa,
        regExp2Nfa,
        eNfa2Nfa, 
        nfa2RegExp, 
        nfa2Dfa,
        nfa2RedNfa,
        dfa2Nfa, 
        dfa2MinDfa,
        dfa2Monoid
    };

    public AbstractAction[] getAllActions() {
        return allActions;
    }
  
    protected void toggleActions() {
        if( source== null ) {
            
            toggleActions( allActions, false );

        } else {

            source.apply( new LangDefObjVisitor() {
                    public Object caseNfa(Nfa n) {
                        // handle epsNfa
                        if( n.isEps() ) {
                            eNfa2Nfa.setEnabled( true );
                        } else {
                            toggleActions( nfaActions, true);
                            eNfa2Nfa.setEnabled( false );
                        }
                        toggleActions( dfaActions, false);
                        toggleActions( regExpActions, false);
                        return null;
                    }
                    public Object caseDfa(Dfa d) {
                        toggleActions( nfaActions, false);
                        toggleActions( dfaActions, true);
                        toggleActions( regExpActions, false);
                        return null;
                    }
                    public Object caseRegExp(RegExp r) {
                        toggleActions( nfaActions, false);
                        toggleActions( dfaActions, false);
                        toggleActions( regExpActions, true);
                        // handle genrex
                        if( r.isGeneralized() )
                            regExp2Nfa.setEnabled( false );
                        else
                            genRegExp2Nfa.setEnabled( false );
                        
                        return null;
                    }
                    public Object caseMonoid(Monoid m) {
                        toggleActions(allActions, false);
                        return null;
                    }
                });
        }
    }
}
