package net.sourceforge.amore.gui.Actions ;

import net.sourceforge.amore.gui.SessionTree.SessionTree;
import net.sourceforge.amore.gui.SessionTree.GuiNode.*;

import net.sourceforge.amore.Ldo.*;
import net.sourceforge.amore.RegLangSet ;


import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/** this class holds action objects that represent unary transformations
 *  that do not preserve the regular language (e.g. L*). is a singleton.
 */
public class TransfUnaryB extends Arrows {

    // singleton methods

    private static TransfUnaryB instance ;
    
    private TransfUnaryB() {
    }
    
    public static TransfUnaryB getInstance() {
        if( instance == null )
            instance = new TransfUnaryB();
        return instance;
    }


    public static SessionTree sessionTree = SessionTree.getInstance();

    public static class STAR_NFA extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    // System.out.println("STAR_NFA"); 
	    Nfa n1 = (Nfa) source;
	    Nfa n2 = TransformsB.starNfa( n1 );
            RegLangSet rset = new RegLangSet(n2);
            rset.setName("star");
            rset.setDescr("L* of language "+n1.getRegSet().getName());
	    sessionTree.addRegLangSet( rset );
	}
	public  STAR_NFA() {
	    super();
	    putValue(Action.NAME, "Nfa recognizing L*");
	    setEnabled(false);
	}
    }
    
    public static STAR_NFA starNfa = new STAR_NFA();
    

    public static class PLUS_NFA extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    // System.out.println("PLUS_NFA"); 
	    Nfa n1 = (Nfa) source;
	    Nfa n2 = TransformsB.plusNfa( n1 );
            RegLangSet rset = new RegLangSet(n2);
            rset.setName("plus");
            rset.setDescr("L+ of language "+n1.getRegSet().getName());
	    sessionTree.addRegLangSet( rset );
	}
	public  PLUS_NFA() {
	    super();
	    putValue(Action.NAME, "Nfa recognizing L+");
	    setEnabled(false);
	}
    }
    
    public static PLUS_NFA plusNfa = new PLUS_NFA();

    public static class REV_NFA extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    // System.out.println("REV_NFA"); 
	    Nfa n1 = (Nfa) source;
	    Nfa n2 = TransformsB.revNfa( n1 );
            RegLangSet rset = new RegLangSet(n2);
            rset.setName("reverse");
            rset.setDescr("reverse of language "+n1.getRegSet().getName());
	    sessionTree.addRegLangSet( rset );
	}
	public  REV_NFA() {
	    super();
	    putValue(Action.NAME, "Nfa recognizing rev(L)");
	    setEnabled(false);
	}
    }
    
    public static REV_NFA revNfa = new REV_NFA();

    public static class SUFF_NFA extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    // System.out.println("SUFF_DFA"); 
	    Dfa d = (Dfa) source;
	    Nfa n = TransformsB.suffNfa( d );
            RegLangSet rset = new RegLangSet(n);
            rset.setName("suffixes");
            rset.setDescr("suffixes of language "+d.getRegSet().getName());
	    sessionTree.addRegLangSet( rset );
	}
	public  SUFF_NFA() {
	    super();
	    putValue(Action.NAME, "Nfa recognizing suff(L)");
	    setEnabled(false);
	}
    }
    public static SUFF_NFA suffNfa = new SUFF_NFA();    

    public static class COMPL_DFA extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    // System.out.println("COMPL_DFA"); 
	    Dfa d1 = (Dfa) source;
	    Dfa d2 = TransformsB.complDfa( d1 );
            RegLangSet rset = new RegLangSet(d2);
            rset.setName("complement");
            rset.setDescr("complement of language "+d1.getRegSet().getName());
	    sessionTree.addRegLangSet( rset );
	}
	public  COMPL_DFA() {
	    super();
	    putValue(Action.NAME, "Dfa recognizing ~L");
	    setEnabled(false);
	}
    }
    public static COMPL_DFA complDfa = new COMPL_DFA();
    
    public static class MINL_DFA extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    // System.out.println("MINL_DFA"); 
	    Dfa d1 = (Dfa) source;
	    Dfa d2 = TransformsB.minL_Dfa( d1 );
            RegLangSet rset = new RegLangSet(d2);
            rset.setName("minL");
            rset.setDescr("minimal words of language "+d1.getRegSet().getName());
	    sessionTree.addRegLangSet( rset );
	}
	public  MINL_DFA() {
	    super();
	    putValue(Action.NAME, "Dfa recognizing min(L)");
	    setEnabled(false);
	}
    }
    public static MINL_DFA minLDfa = new MINL_DFA();    
    
    public static class MAXL_DFA extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    // System.out.println("MAXL_DFA"); 
	    Dfa d1 = (Dfa) source;
	    Dfa d2 = TransformsB.maxL_Dfa( d1 );
            RegLangSet rset = new RegLangSet(d2);
            rset.setName("maxL");
            rset.setDescr("maximal words of language "+d1.getRegSet().getName());
	    sessionTree.addRegLangSet( rset );
	}
	public  MAXL_DFA() {
	    super();
	    putValue(Action.NAME, "Dfa recognizing max(L)");
	    setEnabled(false);
	}
    }
    public static MAXL_DFA maxLDfa = new MAXL_DFA();    

    public static class PREF_DFA extends AbstractAction { 
	public void actionPerformed(ActionEvent e) {
	    // System.out.println("PREF_DFA"); 
	    Dfa d1 = (Dfa) source;
	    Dfa d2 = TransformsB.prefDfa( d1 );
            RegLangSet rset = new RegLangSet(d2);
            rset.setName("prefixes");
            rset.setDescr("prefixes of language "+d1.getRegSet().getName());
	    sessionTree.addRegLangSet( rset );
	}
	public  PREF_DFA() {
	    super();
	    putValue(Action.NAME, "Dfa recognizing pref(L)");
	    setEnabled(false);
	}
    }
    public static PREF_DFA prefDfa = new PREF_DFA();    

        

    public static AbstractAction[] nfaActions = { 
        starNfa, plusNfa, revNfa   };
    public static AbstractAction[] minDfaActions = {
        suffNfa, prefDfa, maxLDfa,
    };
    public static AbstractAction[] dfaActions = { 
        complDfa, minLDfa };

    final static AbstractAction allActions[] = {
        starNfa,
        plusNfa,
        revNfa,
        suffNfa,
        complDfa,
        minLDfa,
        maxLDfa,
        prefDfa,
    };

    public AbstractAction[] getAllActions() {
        return allActions;
    }

    protected void toggleActions() {
        if( source == null ) {
            
            toggleActions( allActions, false );

        } else {
            
            source.apply( new LangDefObjVisitor() {
                    public Object caseNfa(Nfa n) {
                        // handle epsNfa
                        if( n.isEps() ) {
                            toggleActions( allActions, false);
                        } else {
                            toggleActions( nfaActions,  true);
                            toggleActions( dfaActions,  false);
                            toggleActions( minDfaActions, false);
                        }
                        return null;
                    }
                    public Object caseDfa(Dfa d) {
                        toggleActions( nfaActions, false);
                        
                        toggleActions( minDfaActions, d.isMinimal());
                        toggleActions( dfaActions, false);
                        return null;
                    }
                    public Object caseRegExp(RegExp r) {
                        toggleActions(allActions, false);
                        return null;
                    }
                    public Object caseMonoid(Monoid m) {
                        toggleActions(allActions, false);
                        return null;
                    }
                });
        }
    }
}
