/** \file AmoreFrame.java
 */
package net.sourceforge.amore.gui;

import net.sourceforge.amore.gui.global.Resource ;
import net.sourceforge.amore.gui.SessionTree.SessionTree ;
import net.sourceforge.amore.gui.Dialogs.CreateModifyDialog ;
import net.sourceforge.amore.gui.Dialogs.AboutBox ;
import net.sourceforge.amore.gui.Dialogs.NotImplBox ;


import java.awt.*;
import java.awt.event.*;
import javax.swing.* ;
import javax.swing.event.* ;
import javax.swing.border.Border;


/** the application frame. contains AmoreSplitPane with two windows, a
 *  menu and a status line. one window (on the left) is for selecting
 *  LangDefObj objects, one (on the right) is for displaying the
 *  currently selected LangDefObj, menu and status bar.
 *  control flow does not always originate from here.
 *  see package Actions
 *
 */

public class AmoreFrame extends JFrame implements ActionListener {

    boolean DEBUG = Resource.getInstance().getDebug();

    /** the AmoreSplitPane holding the two windows */
    static AmoreSplitPane mainPanel;


    /** layoutmanager.  north south (east west) layout suffices. */
    BorderLayout borderLayout1 = new BorderLayout();

    JMenuBar menuBar1 = new JMenuBar();     // menu bar

    final JMenu menuFile          = FileMenu.getInstance();           
    final JMenu menuTransf        = TransformationMenu.getInstance();
    final JMenu menuTransfBinary  = BinaryTransformationMenu.getInstance();

    /* JMenu menuTest = new JMenu(); */     // to do: test menu
    /* JMenu menuHist = new JMenu(); */     // to do: history menu

    // JMenuItem menuHistShow      = new JMenuItem();
    // JMenuItem menuHistShowLast  = new JMenuItem();

    JMenu menuHelp = new JMenu();              // help menu (A LOT to do)

    JMenuItem menuHelpAbout = new JMenuItem(); 

    static JLabel statusBar = new JLabel();

    public static void setStatusText(String s) {
	// add space
	statusBar.setText(" "+s);
    }

    protected void initAll() {
	JPanel         contentPane = (JPanel) this.getContentPane();

        contentPane.setLayout(borderLayout1);

	// Menu "File"
	JMenuItem menuFileExit = new JMenuItem();
        menuFileExit.setActionCommand("FileExit");
        menuFileExit.addActionListener( this );
	menuFileExit.setText("Exit");
        menuFile.add(menuFileExit);
	
        // Menu "Help"
        menuHelp.setText("Help");
        menuHelpAbout.setText("About");
        menuHelpAbout.setActionCommand("HelpAbout");
        menuHelpAbout.addActionListener( this );
        menuHelp.add(menuHelpAbout);

	// add Menus to menuBar
        menuBar1.add(menuFile);
        menuBar1.add(menuTransf);
        menuBar1.add(menuTransfBinary);
        menuBar1.add(menuHelp);

        /* uncomment this and the decl above to have a debug menu
	if( DEBUG ) {
	    JMenu menuDebug = new JMenu();
	    JMenuItem menuDebugTree = new JMenuItem();
	    menuDebug.setText("Debug");
	    menuDebugTree.setText("DebugTree");
	    menuDebugTree.addActionListener(new GenericAdapter(DEBUG_TREE, this));
	    menuDebug.add(menuDebugTree);        
	    menuBar1.add(menuDebug);
	}
        */


	// status bar
	statusBar.setText("hello !");
	statusBar.setBorder( BorderFactory.createEtchedBorder() );

        contentPane.add(statusBar, BorderLayout.SOUTH);
	contentPane.add( mainPanel,  
			 BorderLayout.CENTER );
                                                                
        contentPane
            .setBorder(BorderFactory
                       .createMatteBorder(10,10,10,10,
                                          new Color(197,197,197)));
                              
        this.setJMenuBar(menuBar1);
    }
    
    // singleton methods
    
    static AmoreFrame instance = null;

    /** called by Main class at startup. This method is the only way
        to construct an AmoreFrame instance */
    public static AmoreFrame createInstance(SessionTree sT) {
	if( instance == null ) {
	    instance = new AmoreFrame( sT );
	}
	return instance;
    }
    
    /** returns the single AmoreFrame instance
     */
    public static AmoreFrame getInstance() {
	if( instance != null )
	    return instance;
	return null;
    }

    /** private constructor. used only by createInstance.
     */
    
    private AmoreFrame(SessionTree sessionTree) {
        // construct a JFrame with the name of the session as title
	super( sessionTree.getName() );

        // create AmoreSplitPane
	mainPanel = AmoreSplitPane.createInstance( sessionTree );
	mainPanel.updateRight();

        // create initialise menus
	initAll();	

        // initialise CreateModifyDialog 
        // (used for creating/modifying LangDefObj objects
	CreateModifyDialog.initialize( this );
        
    }
    

    
    /** overridden so we can exit the program when window is closed
     */
    protected void processWindowEvent(WindowEvent e) {
        super.processWindowEvent(e);
        if (e.getID() == WindowEvent.WINDOW_CLOSING) {
            fileExit();
        }
    }
    
    /** this is called when user wishes to end program execution
     *  To Do: check for changed LangDefObj objects, ask for confirmation
     */
    protected void fileExit() {
	System.exit(0);
    }

    /** delegates update call to AmoreSplitPane (??) To Do: observer
        pattern */
    public static void update() {
	mainPanel.updateRight();
    }

    public void actionPerformed( ActionEvent e ) {
        if( e.getActionCommand().equals("FileExit") )
            fileExit();

        /* not implemented
        if( e.getActionCommand().equals("HistoryShow") )
	    show_NotImplemented_Box();

        if( e.getActionCommand().equals("HistoryShowLast") )
	    show_NotImplemented_Box();
        */
        if( e.getActionCommand().equals("HelpAbout") ) {   // show about box
	    JDialog dlg = new AboutBox(this);
	    showDialog(dlg);
        }
    }

    /** shows a Box saying "Not Implemented" */     
    public void show_NotImplemented_Box() {
        NotImplBox dlg = new NotImplBox(this);
	showDialog(dlg);
    }

    /** convenience method, shows a modal dialog in the middle of the
     *    screen 
     */
    void showDialog(JDialog dlg) {
        Dimension dlgSize = dlg.getPreferredSize();
        Dimension frmSize = getSize();
        Point loc = getLocation();
        dlg.setLocation((frmSize.width  - dlgSize.width)  / 2 + loc.x, 
                        (frmSize.height - dlgSize.height) / 2 + loc.y);
        dlg.setModal(true);
        dlg.show();
    }

}
