/** \file BinaryTransformationMenu.java
 */
package net.sourceforge.amore.gui;

import javax.swing.AbstractAction ;
import javax.swing.JMenu ;
import javax.swing.JMenuItem ;

import net.sourceforge.amore.Ldo.LangDefObj ;
import net.sourceforge.amore.gui.Actions.TransfBinary ;

import net.sourceforge.amore.gui.SessionTree.SessionTree ;
import net.sourceforge.amore.gui.SessionTree.SessionTreeObserver ;
import net.sourceforge.amore.gui.SessionTree.GuiNode.* ;


/** menu containing actions that trigger binary transformations.
 *  see also TransfBinary. is a singleton.
 */
public class BinaryTransformationMenu extends JMenu 
    implements SessionTreeObserver {

    private BinaryTransformationMenu() {
        super();
        setText("bin. Transformation");

        AbstractAction actions[]; 

        actions = TransfBinary.getInstance().getAllActions();

        for(int j=0; j<actions.length; j++)
	    add( actions[j] );

        // register with SessionTree
        SessionTree.getInstance().register( this );
    }
    
    private static BinaryTransformationMenu instance = null;
    
    public  static BinaryTransformationMenu getInstance() {
        if( instance == null )
            instance = new BinaryTransformationMenu();
        return instance;
    }

    /*
    public void toggleAllOff() {
        TransfBinary.toggleActions( TransfBinary
                                    .allActions, 
                                    false);
    }
    */
    
    /** adapts menu by toggling actions, whenever user changes selection.
     */
    public void selectedObjectChanged(GuiNode newSelectedObject) {

        // if new selected object is LwpNode (holds an LangDefObj)
        // get the LangDefObj and pass it as a new source

        LangDefObj ldo = (LangDefObj) newSelectedObject
            .apply(new GuiNodeVisitor() {
                    public Object caseRootNode( RootNode r ) {
                        return null; 
                    }
                    public Object caseRegLangSetNode( RegLangSetNode rn ) {
                        return null; 
                    }
                    public Object caseLwpNode( LwpNode ln ) {
                        return ln.getLdo(); 
                    }
                });

        // setSource allows "null" to be passed. in this case
        // all actions are toggled off
        
        System.out.println( ldo );
        TransfBinary.getInstance().setSource( ldo );
    }
    
}
