package net.sourceforge.amore.gui.Dialogs ;

import net.sourceforge.amore.gui.EditPanels.* ;
import net.sourceforge.amore.gui.Panels.RootPanel ;
import net.sourceforge.amore.gui.global.* ; // for EditMethods,Resource.DEBUG

import net.sourceforge.amore.gui.util.* ;

import net.sourceforge.amore.RegLangSet ;
import net.sourceforge.amore.Ldo.* ;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

/** dialog to create/modify a RegLangSet together with a LangDefObj
 */

public class CreateModifyDialog extends JDialog {

    final int default_sno = 1;        // 1 letter alphabet
    final int default_inputclass = 0; // dfa

    private static boolean DEBUG;

    private static CreateModifyDialog dialog = null;

    private static RegLangSet rset = null;
    private static LangDefObj ldo = null;

    private static LdoEditPanel editPane = null;
    private static JScrollPane scrollPane;
    private static JLabel nameLabel;
    private static JTextField nameField;
    private static JLabel descrLabel;
    //private static JTextField descrField;
    private static JTextArea descrField;
    private static JLabel snoLabel;
    private static WholeNumberField snoField;
    private static JLabel inputByLabel;
    private static JComboBox inputBy;


    
    /** Set up the dialog.  The argument can be null,
     *  but it really should be a component in the dialog's
     *  controlling frame.
     */
    public static void initialize(Component comp) {
	Frame frame = JOptionPane.getFrameForComponent(comp);
	DEBUG = Resource.getInstance().getDebug();
	// scrollPane for editing the ldo
	scrollPane = new JScrollPane();
	dialog = new CreateModifyDialog( frame );
	if( DEBUG ) System.out.println("CreateModifyDialog initialized");
    }

    /**
     * Show the initialized dialog.  The first argument should
     * be null if you want the dialog to come up in the center
     * of the screen.  Otherwise, the argument should be the
     * component on top of which the dialog should appear.
     */
    public static RegLangSet showDialog(Component comp, RegLangSet initialValue) {
	if( DEBUG ) {
	    System.out.println("showDialog called with "+initialValue);
	}
        if (dialog != null) {
	    // display dialog
	    dialog.setValues(initialValue); // if null ->defaults
	    dialog.setLocationRelativeTo(comp);
	    dialog.setVisible(true);
        } else {
            System.err.println("CreateModifyDialog requires you to call "
                               + "initialize before calling showDialog.");
        }
	/*
	if( DEBUG && rset != null ) {
	    System.out.println("CreateModifyDialog finished");
	    System.out.println("result:");
	    System.out.println("Name " + rset.getName());
	    System.out.println("letters " + rset.getNletters());
	    System.out.println("input class "+ rset.getInputClass());
	}
	*/
        return rset;
    }   

    /** create the Bundle object to be returned.
     */
    public void updateValue() {
	rset.setName( nameField.getText() );
	rset.setDescr( descrField.getText() );
	rset.setNletters( snoField.getValue() );
	
	// get clazz from EditMethods via selected index of "inputBy" box
	Class clazz = EditMethods.getClassByIndex( inputBy.getSelectedIndex());
	rset.setInputClass( clazz );	
	LangDefObj input = editPane.getLdo();
	rset.setInput( input ); // might be null
	if( input == null ) {
	    System.out.println("Cre..updateValue: getLdo return null");
	}
    }
    
    public void updateEditPane() {
	Class c = EditMethods.getClassByIndex(inputBy.getSelectedIndex());
	int sno = snoField.getValue();
	rset.setNletters( snoField.getValue() );
	rset.setInputClass( c );
	editPane = EditPanelFactory.createEditPanel( c, sno );
	scrollPane.setViewportView( editPane );
	// RESIZING ???
	//scrollPane.invalidate();
	//getContentPane().invalidate();
	//getContentPane().getLayoutManager().invalidateLayout(scrollPane);
	//getContentPane().add(scrollPane, BorderLayout.CENTER);
    }

    //Overridden so we can exit when window is closed
    protected void processWindowEvent(WindowEvent e) {
        super.processWindowEvent(e);
        if (e.getID() == WindowEvent.WINDOW_CLOSING) {
	    // -> cancel
	    CreateModifyDialog.rset = null;
	    CreateModifyDialog.dialog.setVisible(false);
        }
    }    

    private void setValues( RegLangSet obj ) {
	System.out.println("Cre..setValues");
	if( obj != null) {
	    // modify existing object
	    rset = obj;
	    /* set fields to params */
	    nameField.setText( obj.getName() );
	    descrField.setText( obj.getDescr() );
	    snoField.setValue( obj.getNletters() );
	    int i = EditMethods.index( obj.getInputClass() );
            System.err.println("input class: "+obj.getInputClass()+"  index " +i);
	    inputBy.setSelectedIndex( i );
	    editPane = EditPanelFactory.createEditPanel( obj.getInput() );
	} else {
	    // create new
	    rset = new RegLangSet();
	    nameField.setText("");
	    descrField.setText("");
	    snoField.setValue( default_sno );
	    /* --> EditMethod.inputClasses[ 0 ] is default */
	    int index = 0; // FIXME is default, should appear elsewhere
	    inputBy.setSelectedIndex( default_inputclass );
	    // get clazz from EditMethods via selected index of "inputBy" box
	    Class clazz = EditMethods.getClassByIndex( default_inputclass );
	    rset.setInputClass( clazz );
	    editPane = EditPanelFactory.createEditPanel( clazz, default_sno );
	    
	}
	scrollPane.setViewportView( editPane );
	System.out.println("editPane present : "+(editPane != null));
    }
    
    private CreateModifyDialog(Frame frame) {
	super(frame, "Creating/Modifying Language + LDO", true);	
	
	//buttons
        JButton cancelButton = new JButton("Cancel");
        final JButton setButton = new JButton("Set");
        cancelButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    CreateModifyDialog.rset = null;
		    CreateModifyDialog.dialog.setVisible(false);
		}
	    });
	final CreateModifyDialog q = this;
	final Frame fr = frame;
        setButton.addActionListener(new ActionListener() {
		// 
		// set button pressed !
		//
		public void actionPerformed(ActionEvent e) {
		    q.updateValue();
		    /* FIXME better place for sanity check would be
		       Bundle */
		    System.out.println("setbutton gets value"+rset.toString());
		    if( !rset.isValid() ) {
			JOptionPane.showMessageDialog(fr, "invalid parameters");
		    } else {
			// is ldo validd ?!
			CreateModifyDialog.dialog.setVisible(false);
		    }
		}
	    });
	getRootPane().setDefaultButton(setButton);

        //main part of the dialog

	nameLabel = new JLabel("Name");
	nameField = new JTextField( RegLangSet.getMaxLenName() );
	descrLabel=new JLabel("Description");
	descrField = new JTextArea(2, RegLangSet.getMaxLenDescr()); 
        /* 5 rows, 1 column */
	descrField.setLineWrap(true);
	descrField.setWrapStyleWord(true);
	JScrollPane descrScroll = new JScrollPane(descrField);
	snoLabel = new JLabel("number of letters");
	snoField = new WholeNumberField(5,5);
        snoField.setEditable(false);
	inputByLabel = new JLabel("input obj");
	inputBy = new JComboBox( EditMethods.getInputClassesTerse() );
	//final JCombobox cb = inputBy;
	// Invoked when an item has been selected
	inputBy.addItemListener(new ItemListener() {
		public void itemStateChanged(ItemEvent e) {
		    updateEditPane();
		}
	    });

	nameLabel.setLabelFor(nameField);
	descrLabel.setLabelFor(descrScroll);
	//descrLabel.setLabelFor(descrField);
	snoLabel.setLabelFor(snoField);

	GridBagLayout gridbag = new GridBagLayout();
	GridBagConstraints c = new GridBagConstraints();
	JPanel paramPane = new JPanel();
        paramPane.setLayout( gridbag );
	
	c.gridwidth = 5; c.gridheight = 1;
        JLabel label = new JLabel("Parameters for new language");
	gridbag.setConstraints( label, c );
	paramPane.add( label );

	c.weightx = 1;
	c.weighty = 1;
	c.ipadx = 5; 
	c.ipady = 5;
	c.fill = GridBagConstraints.BOTH;
	c.gridy = 1; // ---------------------------------------------
	LayoutHelp.addSpacer(0,10, paramPane, gridbag, c);
	c.gridy++;
	c.gridx = 0;
	c.gridheight = 1;
	c.gridwidth = 1;
	//c.gridwidth = GridBagConstraints.NONE;
	// paramPane.add(Box.createRigidArea(new Dimension(0,5)));

	gridbag.setConstraints( nameLabel, c );
	paramPane.add( nameLabel );
	c.gridx = 1;
	c.gridwidth = 2;	
	c.weightx = 1;
	gridbag.setConstraints( nameField, c );
	paramPane.add( nameField );

	c.gridy++; // ----------------------------------------------
	LayoutHelp.addSpacer(0,10, paramPane, gridbag, c);
	c.gridy++;
	// number of letters
	c.gridx = 0;
	c.gridwidth = 1;
	gridbag.setConstraints( snoLabel, c );
	paramPane.add( snoLabel );
	c.gridx = 1;
	c.gridwidth = 2;
	gridbag.setConstraints( snoField, c );
	paramPane.add( snoField );
	

	JButton addL = new JButton("add letter");
	
	addL.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                editPane.addLetter();
                snoField.setValue(editPane.getLdo().getNletters());
            }
        });
	
	c.gridx = 4;
	c.gridwidth = GridBagConstraints.RELATIVE; //next to last
	gridbag.setConstraints( addL, c );
	paramPane.add( addL );
	c.gridx = 3;
	JButton remL = new JButton("rem.letter");
	
	remL.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                editPane.remLetter();
                snoField.setValue(editPane.getLdo().getNletters());
            }
        });
	
	gridbag.setConstraints( remL, c );
	paramPane.add( remL );
	
	
	c.gridy++;  // ----------------------------------------
	LayoutHelp.addSpacer(0,10,paramPane, gridbag, c);
	c.gridy++;
	c.gridx = 0;
	c.gridwidth = 1;
	gridbag.setConstraints( descrLabel ,c );
	paramPane.add( descrLabel );
	c.gridx = 1;
	c.gridwidth = 3;
	//descrField.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
	gridbag.setConstraints( descrScroll, c );
	paramPane.add( descrScroll  );
	c.gridy++; // -----------------------------------------
	LayoutHelp.addSpacer(0,10,paramPane, gridbag, c);
	c.gridy++;
	c.gridwidth = 1;
	c.gridx = 0;
	gridbag.setConstraints(  inputByLabel, c );
	paramPane.add( inputByLabel );

	c.gridx = 1;
	c.gridwidth = 2;
	gridbag.setConstraints( inputBy, c );
	paramPane.add( inputBy );	
				
	paramPane.setBorder(BorderFactory.createEmptyBorder(10,10,10,10));
	scrollPane.setBorder(BorderFactory.createEmptyBorder(20,20,20,20));
	scrollPane.setViewportView( RootPanel.getInstance() );
	    

	//Lay out the buttons from left to right.
        JPanel buttonPane = new JPanel();
        buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.X_AXIS));
        buttonPane.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
        buttonPane.add(Box.createHorizontalGlue());
        buttonPane.add(cancelButton);
        buttonPane.add(Box.createRigidArea(new Dimension(10, 0)));
        buttonPane.add(setButton);

	//Put everything together, using the content pane's BorderLayout.
        Container contentPane = getContentPane();
        contentPane.add(paramPane, BorderLayout.NORTH);
	contentPane.add(scrollPane, BorderLayout.CENTER);
        contentPane.add(buttonPane, BorderLayout.SOUTH);
	
        pack();
    }
}
