package net.sourceforge.amore.gui.EditPanels ;

import net.sourceforge.amore.gui.global.Resource ;

import net.sourceforge.amore.MutableLdo.* ;
import net.sourceforge.amore.Ldo.* ;
import net.sourceforge.amore.RegLangSet ;

import java.util.Hashtable ;
import java.util.Enumeration ;

/** contains only static methods. acts like a singleton
 */

public class EditPanelFactory {
    private static boolean DEBUG;
       
    protected EditPanelFactory() {
	// don't need instances of this class
    }

    protected static Hashtable  constrHash= null;

    protected static void init() {
	DEBUG = Resource.getInstance().getDebug();
	
	// construct Constructor Objects
	EditPanelFactory.nfaC = new ConstructNfaPanel();
	EditPanelFactory.dfaC = new ConstructDfaPanel();
	EditPanelFactory.regexpC = new ConstructRegExpPanel();

	constrHash= new Hashtable();

	try {
	    // put them somewhere where we can find them
	    constrHash.put( Class.forName("net.sourceforge.amore.Ldo.Nfa"),   
			    EditPanelFactory.nfaC );
	    constrHash.put( Class.forName("net.sourceforge.amore.Ldo.Dfa"),   
			    EditPanelFactory.dfaC );
	    constrHash.put( Class.forName("net.sourceforge.amore.Ldo.RegExp"),
			    EditPanelFactory.regexpC );
	} catch (ClassNotFoundException e) {
	    System.out.println( e.toString() );
	    throw new RuntimeException("Error, class names not consistent. change and Recompile.");
	}
    }

    protected static ConstructorObj lookup(Class c) {
	if ( constrHash == null ) 
	    init(); // first call creates factory
	
	ConstructorObj cons = ((ConstructorObj) constrHash.get( c ));
	if (cons != null)
	    return cons;
	else {
	    if( DEBUG ) {
		StringBuffer cs = new StringBuffer();
		for( Enumeration e = constrHash.keys(); e.hasMoreElements(); ){
		    cs.append( e.nextElement() + "\n" );
		}
		throw new IllegalArgumentException("I do not know how to construct panels for "+c.toString()+" classes. known classes:\n" + cs);  
	    } else {
		throw new IllegalArgumentException("I do not know how to construct panels for "+c.toString()+" classes.");
	    }
	}
    }

    /** returns EditPanel depending on value of s.
     *  sno - size of alphabet
     */
    public static LdoEditPanel createEditPanel(String s, int sno) {
	try {
	    return createEditPanel( Class.forName( s ), sno );
	} catch (ClassNotFoundException e) {
	    throw new IllegalArgumentException(" there is no class named "+s);
	}
    }
    /**
     *  class (Dfa, Nfa, RegExp)
     *  sno   size of alphabet
     */
    public static LdoEditPanel createEditPanel( Class c, int sno ) {
	return lookup(c).editor(sno);	// voila, enjoy
    }

    /** returns right EditPanel for ldo
     */
    public static LdoEditPanel createEditPanel(LangDefObj ldo) {
	ConstructorObj cons;
	
	if ( constrHash == null ) 
	    init(); // first call creates factory
	Class key = ldo.getClass();
	if( DEBUG )
	    System.out.println("EditPanelFactory:looking up "+key.toString());
	cons = ((ConstructorObj) constrHash.get( key ));
	if( cons != null)
	    return cons.editor( ldo );
	else
	    throw new IllegalArgumentException("I do not know how to construct panels for Class "+key);
	// voila, enjoy
    }
    
    // 
    interface ConstructorObj {
	public LdoEditPanel editor(int sno) ;     // creating
	public LdoEditPanel editor(LangDefObj o) ;// modifying
    }
    static class ConstructNfaPanel implements ConstructorObj {
 	public LdoEditPanel editor(int sno)  {
	    return new MutableNfaPanel(new MutableNfa(sno));
	}
	public LdoEditPanel editor(LangDefObj o) {
	    return new MutableNfaPanel(new MutableNfa( (Nfa) o ));
	}
    }    
    static class ConstructDfaPanel  implements ConstructorObj {
 	public LdoEditPanel editor(int sno)  {
	    return new MutableDfaPanel(new MutableDfa(sno));
	}
	public LdoEditPanel editor(LangDefObj o) {
	    return new MutableDfaPanel(new MutableDfa( (Dfa) o ));
	}
    } 
    static class ConstructRegExpPanel  implements ConstructorObj {
	/* FIXME */
 	public LdoEditPanel editor(int sno) {
	    return new RegExpEditPanel(new RegExp("%",sno));
	}
	public LdoEditPanel editor(LangDefObj o) {
	    return new RegExpEditPanel((RegExp) o);
	}
    } 

    static ConstructNfaPanel    nfaC;
    static ConstructDfaPanel    dfaC;
    static ConstructRegExpPanel regexpC;

}
