package net.sourceforge.amore.gui.EditPanels.Fields ;

import javax.swing.*; 
import javax.swing.text.*; 

import java.awt.Toolkit;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.Locale;

/** JTextField for editing dfa transition targets.
 *  a close relative to util.WholeNumberField.
 */

public class DfaTransitionField extends JTextField {
  private Toolkit toolkit;
  private NumberFormat integerFormatter;
  private boolean qPresent;
  
  private int max;

    /* works only in jdk >= 1.3
      class MyInputVerifier extends InputVerifier {
      public boolean verify(JComponent input) {
      DfaTransitionField tf = (DfaTransitionField) input;
      int target = tf.getValue();
      return (target<getMax());
      }
      }
    */ 
    public DfaTransitionField(int value, int columns, int max) {
        super(columns);
	// see above
	// setInputVerifier( new MyInputVerifier() );
	this.max = max;
	qPresent = false;
        toolkit = Toolkit.getDefaultToolkit();
        integerFormatter = NumberFormat.getNumberInstance(Locale.US);
        integerFormatter.setParseIntegerOnly(true);
        setValue(value);
    }

  public int getMax() {
    return max;
  }

  public void setMax(int max) {
    this.max = max;
  }

    public int getValue() {
        int retVal = 0;
        try {
	    String s = getText();
	    if( qPresent )
		s = s.substring(1);
            retVal = integerFormatter.parse(s).intValue();
        } catch (ParseException e) {
            // This should never happen because insertString allows
            // only properly formatted data to get in the field.
            toolkit.beep();
        }
        return retVal;
    }

    public void setValue(int value) {
        setText(integerFormatter.format(value));
    }

    protected Document createDefaultModel() {
        return new WholeNumberDocument();
    }

    protected class WholeNumberDocument extends PlainDocument {
        public void insertString(int offs, 
                                 String str,
                                 AttributeSet a) 
                throws BadLocationException {
            char[] source = str.toCharArray();
	    char[] result = new char[source.length];
	    int i = 0; // points at source
            int j = 0; // points at result
	    // handle potential 'q' at beginning
	    if(offs==0) {
		qPresent = false;
		char c = source[i]; // first char
		if (c=='q'||c=='Q') {
		    qPresent = true;
		    result[j++] = c; // copy 'q'
		    i++;             // forward one char
		}
	    }
	    // i is either 0 or 1
	    for (; i < result.length; i++) {
		if (Character.isDigit(source[i]))
		    result[j++] = source[i];
                else {
                    toolkit.beep();
                    /* System.err.println("insertString: " + source[i]); */
                }
            }
            super.insertString(offs, new String(result, 0, j), a);
        }
    }
}
