/** \file MutableDfaTableModel.java
 */

package net.sourceforge.amore.gui.EditPanels.Tables ;

import net.sourceforge.amore.gui.util.QTool ;
import net.sourceforge.amore.MutableLdo.MutableDfa ;
import net.sourceforge.amore.Ldo.Dfa ;

import java.awt.* ;
import javax.swing.* ;
import javax.swing.table.* ;

public class MutableDfaTableModel extends AbstractTableModel {
    private static boolean DEBUG = true;

    MutableDfa mdfa;

    public int getColumnCount() { 
	// nletters + 3 (first col = names, 2nd = init, 3rd = final)
	return mdfa.getNletters() + 3;
    }
    public int getRowCount() { 
	return mdfa.getNstates();
    }
    public Class getColumnClass(int i) {
      return getValueAt(0,i).getClass();
    }
    public String getColumnName( int column ) {
	switch (column) {
	case 0:
	    return "";
	case 1:
	    return "init";
	case 2:
	    return "final";
	default:
	}
	// 0 -> "eps", 1 -> "a", ...
	// col - 3(name,init,final) + 1 (no eps letters)
	return QTool.letters[ column - 2];
    }

    /** only names are not mutable.
     */
    public boolean isCellEditable(int row, int col) {
        if (col == 0) 
	    { return false; } 
	else 
	    { return true; } 
    }

    public Object getValueAt(int row, int col) {
	switch( col ) {
	case 0: // first col. contains state-names 
	    return QTool.qQ(row, row == mdfa.getInit(), mdfa.isFinal( row ));
	case 1: // 2nd col -> init state
	    return new Boolean(row == mdfa.getInit());
	case 2: // 3rd col -> final state
	    return new Boolean(mdfa.isFinal( row ));
	default:
	    col = col-2;
	    // delta[ 0 ] unused ( libAMoRE compatibility )
	    // since dfa's don't use epsilons.
	    int i =  mdfa.getcon( row, col );
	    if( i!= MutableDfa.UNDEF )
		return QTool.qQ( i, mdfa.isFinal( i ) );
	    else
		return "?";
	}
    } // getValueAt

    public void setValueAt(Object value, int row, int col) {
	if (DEBUG) {
	    System.out.println("Setting value at " + row + "," + col
			       + " to " + value
			       + " (an instance of " 
			       + value.getClass() + ")");
	}

	switch( col ) {
	case 1: // init
	    if (((Boolean) value).booleanValue() == true) 
		mdfa.setInit( row );
	    else
		mdfa.setInit( MutableDfa.UNDEF );
	    fireTableDataChanged();
	    break;
	case 2:
	    if (((Boolean) value).booleanValue() == true) 
		mdfa.setFinal( row );
	    else
		mdfa.setNonFinal( row );
	    fireTableDataChanged();
	    break;
	default:
	    // value is an integer
	    // ignore
	    System.out.println( value );
	    int target = ((Integer) value).intValue();
	    System.out.println( target );
	    // from state = row
	    // to state = current col - 3(name, init, final) 
	    //                        + 1 (no eps letters, start from 'a')
	    if( target <= mdfa.getNstates() )
		mdfa.connect(row, col - 2, target);
	    else
		// error
		System.out.println("BAD TARGET GIVEN!");
	}
	fireTableCellUpdated(row, col);

	if (DEBUG) {
	    System.out.println("New value of data:");
	    System.out.println(mdfa.toString());
	}
    }

    public Dfa getDfa() {
	// TODO: check if dfa is valid.
	return mdfa.toDfa();
    }

    public MutableDfaTableModel( MutableDfa mdfa ) {
	this.mdfa = mdfa;
    }
}

