/** \file MutableNfaTableModel.java
 */

package net.sourceforge.amore.gui.EditPanels.Tables ;

import net.sourceforge.amore.gui.util.QTool ;
import net.sourceforge.amore.MutableLdo.MutableNfa ;
import net.sourceforge.amore.Ldo.Nfa ;

import java.awt.* ;
import javax.swing.* ;
import javax.swing.table.* ;
import java.util.Vector ;
import java.util.Enumeration ;

public class MutableNfaTableModel extends AbstractTableModel {
    private static boolean DEBUG = true;

    MutableNfa mnfa;

    public int getColumnCount() { 
	// nletters + 3 (first col = names, 2nd = init, 3rd = final)
	return mnfa.getNletters() + 3;
    }
    public int getRowCount() { 
	return mnfa.getNstates();
    }
    public Class getColumnClass(int i) {
      return getValueAt(0,i).getClass();
    }
    public String getColumnName( int column ) {
	switch (column) {
	case 0:
	    return "";
	case 1:
	    return "init";
	case 2:
	    return "final";
	default:
	}
	// 0 -> "eps", 1 -> "a", ...
	// col - 3(name,init,final) + 1 (no eps letters)
	return QTool.letters[ column - 2];
    }

    /** only names are not mutable.
     */
    public boolean isCellEditable(int row, int col) {
        if (col == 0) 
	    { return false; } 
	else 
	    { return true; } 
    }

    public Object getValueAt(int row, int col) {
	switch( col ) {
	case 0: // first col. contains state-names 
	    return QTool.qQ(row, mnfa.isInit( row ), mnfa.isFinal( row ));
	case 1: // 2nd col -> init state
	    return new Boolean(mnfa.isInit( row ));
	case 2: // 3rd col -> final state
	    return new Boolean(mnfa.isFinal( row ));
	default:
	    col = col-2;
	    // row = state, col = letter
	    return QTool.nfaQs( mnfa.getcon( row, col ));
	}
    } // getValueAt

    public void setValueAt(Object value, int row, int col) {
	if (DEBUG) {
	    System.out.print("Setting value at " + row + "," + col
			       + " to " + value );
	    if (value!=null) 
		System.out.println(" (an instance of " 
				   + value.getClass() + ")");
	    else
		System.out.println();	
	}

	switch( col ) {
	case 1: // init
	    if (((Boolean) value).booleanValue() == true) 
		mnfa.setInit( row );
	    else
		mnfa.setNonInit( row );
	    fireTableDataChanged();
	    break;
	case 2:
	    if (((Boolean) value).booleanValue() == true) 
		mnfa.setFinal( row );
	    else
		mnfa.setNonFinal( row );
	    fireTableDataChanged();
	    break;
	default:
	    // value is a vector of integer

	    // from state = row
	    // to state = current col - 3(name, init, final) 
	    //                        + 1 (no eps letters, start from 'a')
	    int letter = col - 3 + (mnfa.isEps() ? 0 : 1);

	    System.out.println( value );
	    Enumeration e=((Vector) value).elements();
	    int i=0;
	    int qno = mnfa.getNstates();
	    while( i<=qno && e.hasMoreElements()) {
		int nextTarget = ((Integer) e.nextElement()).intValue();
		while((i != nextTarget) && (i <=qno)) {
		    mnfa.disconnect( row, letter, i );
		    i++;
		}
		if ( i == nextTarget ) {
		    mnfa.connect( row, letter, i );
		    i++;
		} 
	    }

	    /*
	    if( target <= mnfa.getNstates() )
		mnfa.connect(row, col - 2, target);
	    else
		// error
		System.out.println("BAD TARGET GIVEN!");
	    */
	}
	fireTableCellUpdated(row, col);

	if (DEBUG) {
	    System.out.println("New value of data:");
	    System.out.println(mnfa.toString());
	}
    }

    public Nfa getNfa() {
	// TODO: check if nfa is valid.
	return mnfa.toNfa();
    }

    public MutableNfaTableModel( MutableNfa mnfa ) {
	this.mnfa = mnfa;
    }
}

