package net.sourceforge.amore.gui.Panels ;

import net.sourceforge.amore.gui.global.Resource ;
import net.sourceforge.amore.Ldo.* ;

import java.util.Hashtable ;

/** creates the right panels for a given ldo. see createLdoPanel for
 usage. is a singleton class. */

public class PanelFactory implements LangDefObjVisitor {

    private static PanelFactory instance = null;

    private PanelFactory() {}

    /** returns the PanelFactory. Singleton method.
     */
    public static PanelFactory getInstance() {
	if (instance==null) {
	    instance = new PanelFactory();
	}
	return instance;
    }

    /** creates a panel for the argument ldo
     *  @param  ldo the LangDefObj which needs a panel
     *  @return LangDefObjPanel for the ldo
     */
    public static LangDefObjPanel createLdoPanel(LangDefObj ldo) {
	return (LangDefObjPanel) (ldo.apply( getInstance() ));
    }
    
    /** visitor method */
    public Object caseNfa(Nfa n) {
	return new NfaPanel( n );
    }
    
    /** visitor method */
    public Object caseDfa(Dfa d) {
	return new DfaPanel( d );
    }
    
    /** visitor method */
    public Object caseRegExp(RegExp r) {
	return new RegExpPanel( r );
    }
    
    /** visitor method */
    public Object caseMonoid(Monoid m) {
	return new MonoidPanel( m );
    }
    
}
