package net.sourceforge.amore.gui.SessionTree ;

import net.sourceforge.amore.Ldo.LangDefObj ;
import net.sourceforge.amore.RegLangSet ;
import net.sourceforge.amore.gui.global.Resource;

import net.sourceforge.amore.gui.SessionTree.GuiNode.* ;

import net.sourceforge.amore.gui.AmoreFrame ; // TEMP
import java.util.Vector ;
import java.util.Enumeration ;
import java.util.Iterator ;
import javax.swing.JTree ;
import javax.swing.tree.* ; 

import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeSelectionEvent;


/** holds all objects of the session and is responsible of drawing them.
 */
public class SessionTree extends JTree {
    static Resource      res;

    SessionTreeModel treeModel; /* FIXME make AmoreSession a TreeModel */

    private GuiNode lastNode;
    private GuiNode currentNode;

    /** returns rootnode
     */
    public static GuiNode getRoot() {
	return RootNode.getInstance(); 
    }
    
    /** returns name of the session
     */
    public String getName() {
	return RootNode.getInstance().toString();
    }

    // observer methods
    
    /** the SessionTreeObserver objects that want to be notified on change
     */
    private Vector sessionTreeObservers = new Vector();

    /** by registering, a SessionTreeObserver expresses its wish to be 
     *  notified, when the user changes the selected session tree node
     */
    public void register( SessionTreeObserver o ) {
        sessionTreeObservers.add( o );
    }

    /** by unregistering, a SessionTreeObserver signals that it is no longer
     *  interested in receiving notifications.
     */
    public void unregister( SessionTreeObserver o ) {
        sessionTreeObservers.remove( o );
    }

    /** notifies the observers, called by TreeSelectionListener
     */
    public void notifyObservers() {

        // get currently selected object
        GuiNode newSelectedObject = (GuiNode) getLastSelectedPathComponent(); 
        
        // notify observers
        for (Iterator i=sessionTreeObservers.iterator(); i.hasNext() ; ) {
            ((SessionTreeObserver) i.next())
                .selectedObjectChanged( newSelectedObject );
        }
    }
    
    /** returns the currently selected GuiNode object
     */
    public GuiNode getCurrent() {
	GuiNode n = (GuiNode) getLastSelectedPathComponent();
	if( n == null )
	    return RootNode.getInstance();
	else
	    return n;
    }

    /** returns a Vector of all RegLangSet objects present in this session.
     *  They may not be altered, since changes would not carry over to
     *  their graphical representations
     */
    public Vector getLanguages() {
        // children() of rootnode is enum of RegLangSetNodes
        Vector rs = new Vector();
        for(Enumeration e = getRoot().children(); e.hasMoreElements();)
            rs.add( ((RegLangSetNode) e.nextElement()).getRegLangSet() );
        return rs;
    }

    /** add a RegLangSet object to this session.
     *  a RegLangSetNode object is automatically created
     */
    public void addRegLangSet( RegLangSet r ) {
	treeModel.addRegLangSet( r );
    }

    /** removes the selected object from the session
     */
    public void removeCurrent() {
	final GuiNode node = (GuiNode)
		    getLastSelectedPathComponent();

	node.apply( new GuiNodeVisitor() {
		public Object caseRootNode(RootNode r) {
		    System.out.println("removing whole sessions not impl.ed");
		    return null;
		}
		public Object caseRegLangSetNode(RegLangSetNode b) {
		    if( treeModel.removeRegLangSetNode( b ) ) {
			if( lastNode != null ) {
			    currentNode = null;
			    select( lastNode );
			} else {
			    currentNode = null;
			    select( null );
			}
		    }
		    return null;
		}
		public Object caseLwpNode(LwpNode l) {
		    System.out.println("removing ldos not impl.ed ");
		    return null;
		}
	    });
    }

    public static SessionTree createInstance() {
      //System.out.println("SessionTree.createInstance");
      if( instance != null )
          throw new RuntimeException("attempt to init me twice");
      instance = new SessionTree();
      return instance;
    }
    public static SessionTree getInstance() {
        if( instance == null )
            throw new RuntimeException("not properly initialized");
        return instance;
    }
    
    /*
    public void updateLast() {
        // this is intended to make selection of the second object
        //  for binary Transformations faster. Is not used at the moment

	
	GuiNode n = (GuiNode)getLastSelectedPathComponent();
	if( n != null ) {
	    lastNode    = currentNode;
	    currentNode = (GuiNode) n;
	    System.out.println("XX current: "+currentNode);
	    System.out.println("XX last: "+lastNode);
	}
	
    }
    */

    public void select( GuiNode n ) {
	if( n!=null ) 
	    setSelectionPath( new TreePath( n.getPath() ));
	else
	    setSelectionPath( new TreePath( RootNode.getInstance() ));
    }

    private static SessionTree instance = null;

    private SessionTree() {
	super();

	System.out.println("init'ing SessionTree()");
	treeModel = new SessionTreeModel();
	setModel( treeModel );

	// up to one obj may be selected
	getSelectionModel().setSelectionMode
	    (TreeSelectionModel.SINGLE_TREE_SELECTION);
	
	// should we show the name of the session ?
	res = Resource.getInstance();
	setRootVisible( res.getShow_session_name() );

        // notify observers on selection change
	addTreeSelectionListener( new TreeSelectionListener() {
		public void valueChanged(TreeSelectionEvent e) {
                    notifyObservers();
                }
            });
    }
    
    public void debugTry() {
	System.out.println("debugTry");
	treeModel.insertDEBUG();
    }
}
