package net.sourceforge.amore.gui.SessionTree;

import net.sourceforge.amore.RegLangSet ;
import net.sourceforge.amore.LibExamples ; // FIXME
import net.sourceforge.amore.Ldo.LangDefObj ;
import net.sourceforge.amore.gui.SessionTree.GuiNode.* ;


import javax.swing.event.*;
import javax.swing.tree.*;
import java.util.*;


public class SessionTreeModel implements TreeModel, RegLangSetNodeObserver {
    
    private Vector treeModelListeners = new Vector();

    RootNode root;

    public SessionTreeModel() {

	System.out.println("init'ing SessionTreeModel()");

	root = RootNode.createInstance("LibExamples");
	
	// FIXME example debug code, uses LibExamples
	Vector languages = new Vector( new TreeSet( LibExamples
						    .getInstance()
						    .getRegLangSetVector()));
	for(Enumeration e = languages.elements(); e.hasMoreElements(); ) {
	    RegLangSetNode rn = root
                .addRegLangSet( (RegLangSet) e.nextElement() );
            rn.register( this );
	}
	
    }
    
    public boolean addRegLangSet( RegLangSet r ) {
	RegLangSetNode rn = root.addRegLangSet( r );

	if( rn == null ) {
	    System.out.println("something went wrong with addRegLangSet");
	    return false;
	}

        // register to get notified when an lwpNode is added
        rn.register( this );        

	int i = root.getIndex(rn);
	fireTreeNodesInserted(root.getPath(),
			      new int[] { i }, 
			      new Object[] { root, rn });
	return true;
    }
    
    /** this method is called when one of the RegLangSetNode objects in
     *  the tree has a new child */
    public void lwpNodeAdded( RegLangSetNode rn, LwpNode ln ) {
        System.out.println("HELLO lwpNodeAdded");
        int i = rn.getIndex( ln );
	fireTreeNodesInserted(rn.getPath(),
			      new int[] { i }, 
			      new Object[] { root, rn, ln });
    }
    
    public boolean removeRegLangSetNode( RegLangSetNode rn ) {
        rn.unregister( this );
	int i = root.removeRegLangSetNode( rn );
	if( -1 == i ) {
	    System.out.println("something went wrong with removeRegLangSetNode");
	    return false;
	}
	fireTreeNodesRemoved( root.getPath(),
			      new int[] { i }, 
			      new Object[] { root, rn });
	return true;
    }

    public void insertDEBUG() {
	// used for debugging
    }

//////////////// Fire events //////////////////////////////////////////////
    // path : TreePath to parent
    // childIndices, children : see TreeModelEvent
    
    protected void fireTreeNodesInserted(TreePath path,
					 int[] childIndices,
					 Object[] children) {
	System.out.println("firing TreeNodesInserted event");
	TreeModelEvent e = new TreeModelEvent(this, path, 
					      childIndices, children);
	int len = treeModelListeners.size();
	for (int i = 0; i < len; i++) {
            ((TreeModelListener)treeModelListeners.elementAt(i)).
		treeNodesInserted(e);
        }
    }

    protected void fireTreeNodesRemoved(TreePath path,
					 int[] childIndices,
					 Object[] children) {
	System.out.println("firing TreeNodesRemoved event");
	TreeModelEvent e = new TreeModelEvent(this, path, 
					      childIndices, children);
	int len = treeModelListeners.size();
	for (int i = 0; i < len; i++) {
            ((TreeModelListener)treeModelListeners.elementAt(i)).
		treeNodesRemoved(e);
        }
	// System.gc(); ??? 
    }


//////////////// TreeModel interface implementation ///////////////////////


    /**
     * Adds a listener for the TreeModelEvent posted after the tree changes.
     */
    public void addTreeModelListener(TreeModelListener l) {
        treeModelListeners.addElement(l);
    }

    /**
     * Returns the child of parent at index index in the parent's child array.
     */
    public Object getChild(Object parent, int index) {
	GuiNode g = (GuiNode) parent;
	return g.getChildAt( index );
    }

    public int getChildCount(Object parent) {
	return ((GuiNode) parent).getChildCount();
    }

    
    /**
     * Returns the index of child in parent.
     */
    public int getIndexOfChild(Object parent, Object child) {
	GuiNode g = ((GuiNode) parent);
	return g.getIndex((GuiNode) child);
    }
    
    /**
     * Returns the root of the tree.
     */
    public Object getRoot() {
	return (Object) root;
    }
    
    /**
     * Returns true if node is a leaf.
     */
    public boolean isLeaf(Object node) {
	GuiNode g = (GuiNode) node;
	return g.isLeaf();
    }

    /**
     * Removes a listener previously added with addTreeModelListener().
     */
    public void removeTreeModelListener(TreeModelListener l) {
        treeModelListeners.removeElement(l);
    }
    /**
     * Messaged when the user has altered the value for the item
     * identified by path to newValue.  Not used by this model.
     */
    public void valueForPathChanged(TreePath path, Object newValue) {
        System.out.println("*** SessionTreeModel: valueForPathChanged : "
                           + path + " --> " + newValue);
    }
}
