/** \file TransformationMenu.java
 */
package net.sourceforge.amore.gui;

import javax.swing.AbstractAction ;
import javax.swing.JMenu ;
import javax.swing.JMenuItem ;

import net.sourceforge.amore.Ldo.LangDefObj ;
import net.sourceforge.amore.gui.Actions.TransfUnaryA ;
import net.sourceforge.amore.gui.Actions.TransfUnaryB ;

import net.sourceforge.amore.gui.SessionTree.SessionTree ;
import net.sourceforge.amore.gui.SessionTree.SessionTreeObserver ;
import net.sourceforge.amore.gui.SessionTree.GuiNode.* ;

/** menu containing actions that trigger unary transformations.
 * see also TransfUnaryA, TransfUnaryB. is a singleton.
 */
public class TransformationMenu extends JMenu 
    implements SessionTreeObserver {

    private TransformationMenu() {
        super();
        setText("Transformation");
        AbstractAction actions[]; 

        // language preserving
        actions = TransfUnaryA.getInstance().getAllActions();        

        for(int j=0; j<actions.length; j++)
	    add( actions[j] );

        addSeparator();
        
        // not language preserving
	actions = TransfUnaryB.getInstance().getAllActions();	

	for(int j=0; j<actions.length; j++)
	    add( actions[j] );

        // register with SessionTree
        SessionTree.getInstance().register( this );
    }
    
    private static TransformationMenu instance = null;
    
    public  static TransformationMenu getInstance() {
        if( instance == null )
            instance = new TransformationMenu();
        return instance;
    }    


    public void selectedObjectChanged(GuiNode newSelectedObject) {

        // if new selected object is LwpNode
        // than adjust menu
        LangDefObj ldo = (LangDefObj) newSelectedObject
            .apply(new GuiNodeVisitor() {
                    
                    public Object caseRootNode( RootNode r ) {
                        return null;
                    }
                    public Object caseRegLangSetNode( RegLangSetNode rn ) {
                        return null;
                    }
                    public Object caseLwpNode( LwpNode ln ) {
                        return ln.getLdo();
                    }
                });
        TransfUnaryA.getInstance().setSource( ldo );
        TransfUnaryB.getInstance().setSource( ldo );
    }
}
