package net.sourceforge.amore.gui.util ;

import java.io.StringReader;
import java.io.IOException;
import java.util.Vector;
import java.util.Enumeration ;

import java.util.TreeSet ;
public class QTool {

    //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    //                    int -> String mappings
    //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    

    /** mapping 0 -> eps, 1 -> a, 2 -> b ...
     */
    public static final String[] letters = new String[] {
	"@", "a", "b", "c", "d", "e", "f", "g", "h", "i", "j",
	    "k", "l", "m", "n", "o", "p", "q", "r", "s", "t", "u",
	    "v", "w", "x", "y", "z" } ;
    
    /** mapping 0 -> "q0", 1 -> "q1" ...
     */
    protected static String q(int i) {
	return new String("q"+i);
    }

    /** mapping 0 -> "Q0", 1 -> "Q1" ...
     */
    protected static String Q(int i) {
	return new String("Q"+i);
    }
    
    /** prefix '->' for init states, capital for final states
     */
    public static String qQ(int i, boolean fin) {
	if( fin )
	    return Q(i);
	else
	    return q(i);
    }

    public static String qQ(int i, boolean ini, boolean fin) {	
	StringBuffer res = new StringBuffer("");
	if( ini )
	    res.append("-> ");
	res.append( qQ( i, fin ) );
	return new String(res);
    }

    //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    //  Vector<Int> -> String mappings
    //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    /** returns string representation of a nfa transition.
     *  see nfa.getcon( F, L ) for format
     */
    public static String nfaQs(Vector v) {
	StringBuffer res = new  StringBuffer("");
	Enumeration e = v.elements();
	for( ; e.hasMoreElements(); ) {
	    Vector vv = (Vector) e.nextElement();
	    // state number
	    int i     = ((Integer) vv.elementAt( 0 )).intValue();
	    // boolean j = ((Boolean) vv.elementAt( 1 )).booleanValue();
	    // final flag
	    boolean k = ((Boolean) vv.elementAt( 2 )).booleanValue();
	    res.append(QTool.qQ(i, k)+",");
	}
	int l = res.length(); // last index.
	if(l>0) // non-empty
	    res.setLength(l-1);
	return res.toString(); 
    }

    //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    //   functions to parse dfa target state / nfa target state list 
    //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    final static char EOF = (char) -1;
    
    /** takes a string (e.g "q3") and returns the target state as an int.
     *   omitting the prefix 'q' or 'Q' is allowed. 
     *   on empty input, -1 is returned
     *   on error, throws IllegalArgumentException
     */

    /*
    public static int dfaParse(String s) {
	int res;
	char c = s.charAt(0);
	try{
	    if(( c=='q' )||( c=='Q' ))
		res = Integer.valueOf(s.substring(1)).intValue();
	    else 
		// we allow omitting 'q' or 'Q' at beginning
		res = Integer.valueOf(s).intValue();
	    if (res<0) {
		throw new IllegalArgumentException("string contains negative num");
	    }
	} catch(StringIndexOutOfBoundsException e) {
	    res = -1;
	}
	return res;
    }
    */

    
    public static boolean isdigit(char c) {
	return (c>='0')&&(c<='9');
    }

    /** takes a string (e.g "q3") and returns the list of target states 
     *   as an Vector (list) of ints.
     *   on empty input, the empty Vector (list) is returned
     *   on error, throws IllegalArgumentException
     */

    public static Vector nfaParse(String s) throws IllegalArgumentException {
	StringReader sr = new StringReader(s);
	Vector v = new Vector();
	char c;

	//DEBUG
	System.out.println("nfaParse(\""+s+"\"");

	try {
	    c = (char) sr.read();
	} catch(IOException e) {
	    throw new IllegalArgumentException(e.toString());
	}
	if (c==EOF) return v; // empty list

	int state = 0;
	int num = -1; 
	
	// num == -1 means invalid input

	boolean spaces = true;
	boolean commata = true;

	while( !(c==EOF) ) {
	    switch(state) {
		// state 0: expect q or Q and nothing else
	    case 0: 
		if(( c=='q' )||( c=='Q' )) {
		    state = 1;
		} else 
		    throw new IllegalArgumentException("not nfa format");
		break;
		// state 1: expect a digit or a comma and nothing else
	    case 1: if( isdigit(c) ) {
		if(num == -1)
		    num = ((int) c - (int) '0');
		else
		    num = num*10 + ((int) c - (int) '0');
	    } else if(( c == ','&&commata )||( c==' ' && spaces )) {
		switch(c) {
		case ',': commata = true; spaces = false; break;
		case ' ': commata = false; spaces = true; break;
		}
		v.add(new Integer(num)); // store last target
		num = -1;                // set to invalid 
		state = 0;
	    } else
		throw new IllegalArgumentException("not nfa format");
	    break;
	    default:
		// this can't happen, we're either 0 or 1
	    }
	    // fetch next char
	    try {
		c = (char) sr.read();
	    } catch(IOException e) {
		throw new IllegalArgumentException(e.toString());
	    }
	}
	if(num != -1)
	    // add last target
	    v.add(new Integer(num));
	else
	    throw new IllegalArgumentException("not nfa format");
	// sort v
	TreeSet ts = new TreeSet(v);
	v = new Vector( ts );
	return v;
    }

    /* this code demonstrates the use of dfaParse, nfaParse ...

    public static void main(String args[]) {
	QString q;
	System.out.println("Hello");
	try {
	    q = new QString(System.in);
	    System.out.println("you entered :"+q);
	    
	    System.out.print("q.dfa :");
	    try {
		int i = q.dfaParse(q.toString());
		if (i == -1) {
		    System.out.println("empty!");
		} else {
		    System.out.println(i);
		}
	    } catch(IllegalArgumentException e) {
		System.out.println("error - not a dfa target");
	    }
	    System.out.print("q.nfa :");
	    try {
		Vector v = q.nfaParse(q.toString());
		if(v.size() == 0) {
		    System.out.println("empty!");
		} else {
		    System.out.println(v);
		}
	    } catch(IllegalArgumentException e) {
		System.out.println("error - not a nfa target");	
	    }
	} catch(Exception e) {
	    System.err.println(e);
	}
    }
    */
}
