#ifndef _TASK_H
#define _TASK_H
/* task.h
 *
 * similar to monads ( cf. Wadler "The Essence of functional programming",
 * Proceedings of 19th POPL ). Please, don't try this at home.       Burak
 */

#include "obj.h"   /* typedef objType */
#include "atime.h" /* clock_getTime() */

struct task_struct {
    /** time this task has used */
    double time_used;
    /** pointer to result */
    int res; 
    /** type of the result */
    objType resType;
    /** body of the task */
    int (*func)(int);
    /** next task in list */
    struct task_struct *next;
} ;

typedef struct task_struct *task;

/** allocate mem for a new task and set body
 */
task maketask(int (*func)(int)) ; 

/** frees all intermediate results of a task.
 *  safe to call with NULL argument.
 *  last result is not freed.
 */
void task_freeresults(task t);

/** execute task(s)
 */
void task_execAll(task t, int firstArg) ;

/** execute task(s), measure time used
 */
void task_execAllTimed(task t, int firstArg) ;

/** append task(s) t2 to task t1 (t1 should be last task!)
 */
void task_append( task t1, task t2 ) ;

/** get result
 */
int task_getRes( task t ) ; 

#endif
