/* Dfa. java
 *
 * Dez. 2000 Burak Emir
 */

package net.sourceforge.amore.Ldo;

import net.sourceforge.amore.Helper.ArrayFunc;

/** a deterministic finite state automaton, maybe minimal */

public class Dfa extends FiniteAutom  {

    static native private int  dfauto_construct( int qno, int sno );
    static native private int  dfauto_clone( int p );
    static native private void dfauto_destruct( int p );

    static native private void    dfauto_setFinal( int p, int q );
    static native private void    dfauto_setNonFinal( int p, int q );
    static native private void    dfauto_setInit( int p, int q );
    static native private boolean dfauto_isFinalState(int p, int q );
    static native private int     dfauto_getInit( int p ) ;

    static native private void   dfauto_connect( int p, int F, int L, int T ) ;
    static native private void   dfauto_disconnect( int p, int F, int L ) ;

    static native private int     dfauto_getcon( int p, int F, int L ) ;
    static native private int     dfauto_getNletters( int p ) ;
    static native private int     dfauto_getNstates( int p ) ;
    static native private void    dfauto_setMin( int p, boolean isred );
    static native private boolean dfauto_isMin( int p );


    /** construct Dfa.
     *  @param nletters size of alphabet
     *  @param nstates  number of states
     */
    public Dfa( int nletters, int nstates ) {
	super( dfauto_construct( nstates - 1, nletters ) );
    }

    /** construct with pointer to C-level Dfa
     */
    public Dfa( int p ) {
	super( p );
    }

    /** if isMinimal() returns MinDfa, otherwise returns "Dfa"
     */
    public String prefix() {
	if(isMinimal()) 
	    return "MinDfa";
	return "Dfa";
    }

    protected void finalize() throws Throwable {
	dfauto_destruct( p );
	super.finalize();
    }

    /** clone method */
    public Object clone() {
	return new Dfa( dfauto_clone( p ) ); // C - level clone
    }

    // ----------------------------------      methods for #states, #letters

    /** returns size of the alphabet (number of letters)                    */
    public int getNletters()               { return dfauto_getNletters( p ); }

    /** returns number of states                                            */
    public int getNstates()                 { return dfauto_getNstates( p ); }


    // ---------------------------------    methods for final/nonfinal states

    /** mark a state as final state */ 
    public void setFinal( int q )              { dfauto_setFinal( p, q );    }


    /** mark a state as nonfinal state */
    public void setNonFinal( int q )           { dfauto_setNonFinal( p, q ); }


    /** is state q a final state ?
     */
    public boolean isFinalState( int q ) {return dfauto_isFinalState( p, q ); }


    // ----------------------------------           methods for initial state

    /** mark a state as an initial state */
    public void setInit( int q )               { dfauto_setInit( p, q );     }

    /** returns the initial state of this dfa
     */
    public int getInit()                       { return dfauto_getInit( p ); }    
    /** returns true if state q is an initial state.
     */
    public boolean isInitState( int q )        { return ( q == getInit() );  }


    // -------------------------------------          methods for transitions

    /** make transition from F to T with Letter L */
    public void connect( int F, int L, int T ){ dfauto_connect( p, F, L, T ); }

    /** destroy transition from F to T */
    public void disconnect( int F, int L )    { dfauto_disconnect( p, F, L ); }
    
    /** getcon: return state to which F is connected via L */
    public int getcon( int F, int L )      { return dfauto_getcon( p, F, L ); }


    // ------------------------------------          minimal flag

    /** returns true if the minimal flag is set
     */
    public boolean isMinimal()                   { return dfauto_isMin( p ); }

    /** sets the minimal flag. this method usually should only be
     *  called with argument "true" if the object is really a minimal
     *  dfa.  */
    public void setMinimal( boolean isMin )     { dfauto_setMin( p, isMin ); }

    // ---- methods for LangDefObjVisitor


    /** calls the caseDfa method of the visitor
     */
    public Object apply( LangDefObjVisitor v ) {
        return v.caseDfa( this );
    }


}
