/* LangDefObj.java
 *
 * Dez. 2000 Burak Emir
 */

package net.sourceforge.amore.Ldo ;

import net.sourceforge.amore.RegLangSet ;

/** superclass of all language defining objects */

public abstract class LangDefObj {

    // ------ attributes

    /** via this pointer one can access the other ldo objects that
     *  recognize the same language, provided they have been computed
     *  already.  */
    protected RegLangSet regSet;
    
    /** pointer to the underlying C-level data structure
     */
    protected int p;

    // ------ constructors

    /** construct a LangDefObj with a pointer to a C data structure
     */
    protected LangDefObj( int p ) {
	this.p = p;
    }

    // ------ navigators

    /** sets the RegSet object, that this object belongs to.
     */
    public void setRegSet( RegLangSet regSet ) {
	this.regSet = regSet;
    }

    /** returns the RegSet object, that this object belongs to.  may
     *  be null, if the RegSet object for this object has not been set
     *  yet. */
    public RegLangSet getRegSet() {
	return regSet;
    }

    /** returns the name of the RegSet that this object belongs to,
     *  "-unnamed-" if this object does not belong to any RegSet */
    public String getName() {
	if( regSet != null )
	    return regSet.getName(); // toString();
	return "-unnamed-";
    }


    /** needed by RegLangSet */
    public int getCpointer() {
	return p;
    }
    

    // --- abstract methods
    
    /** returns Dfa for dfa, Nfa for nfa, RegExp for regexp */
    public abstract String prefix() ;

    /** returns prefix + name of the RegLangSet object */
    public String getFullName() {
	return prefix() + " " + getName();
    }

    public abstract Object clone();


    // ----------------------------------

    /** returns number of letters of the alphabet
     */
    public abstract int getNletters() ; 
    

    // ----------------------------------

    /** same as getFullName.
     */
    public String toString() {
	return getFullName();
    }

    /** override this method to make LangDefObjVisitor work
     */
    public abstract Object apply( LangDefObjVisitor v ) ;

}

