/* Nfa. java
 *
 * Dez. 2000 Burak Emir
 */

package net.sourceforge.amore.Ldo;

import java.util.Vector ;
/** adapter around C "objects". delegates
 */

public class Nfa extends FiniteAutom {
    
    static native private int     nfauto_construct( int qno, int sno, boolean eps );
    static native private int     nfauto_clone( int p );
    static native private void    nfauto_destruct( int p );

    static native private void    nfauto_setFinal( int p, int q );
    static native private void    nfauto_setNonFinal( int p, int q );
    static native private void    nfauto_setInit( int p, int q );
    static native private void    nfauto_setNonInit( int p, int q );
    static native private boolean nfauto_isFinalState(int p, int q );
    static native private boolean nfauto_isInitState(int p, int q ) ;
    static native private void    nfauto_connect( int p, int F, int L, int T ) ;
    static native private void    nfauto_disconnect( int p, int F, int L, int T) ;
    static native private boolean nfauto_testcon( int p, int F, int L, int T ) ;
    static native private int     nfauto_getNletters( int p ) ;
    static native private int     nfauto_getNstates( int p ) ;
    static native private void    nfauto_setEps( int p, boolean iseps );
    static native private boolean nfauto_isEps( int p );
    static native private void    nfauto_setRed( int p, boolean isred );
    static native private boolean nfauto_isRed( int p );
    
    // ------- constructors
    
    /** construct Nfa or EpsNfa. to construct an EpsNfa correctly, 
     *  eps must be true and the size of the alphabet must include all
     *  letters plus 1 for the epsilon.
     *  @param nletters size of alphabet
     *  @param nstates  number of states
     *  @param eps      contains epsilon transitions.
     */
    public Nfa( int nletters, int nstates, boolean eps ) {
	super( nfauto_construct( nstates - (eps?1:0), nletters,  eps ));
	setEps( eps );
    }

    /** construct Nfa. ( no EPS !, not reduced)
     *  @param nletters size of alphabet
     *  @param nstates  number of states
     */
    public Nfa( int nletters, int nstates ) {
	super( nfauto_construct( nstates - 1, nletters,  false ) );
    }

    /** returns an epsilon-Nfa. Here, nletters means the size of
     *  the alphabet without the epsilon.
     *  @param nletters size of alphabet, without epsilon
     *  @param nstates  number of states
     */
    public static Nfa constructEpsNfa( int nletters, int nstates ) {
	return new Nfa(  nletters + 1, nstates, true );
    }

    /** construct with pointer to C-level Nfa.
     */
    public Nfa( int p ) {
	super( p );
    };

    /** returns Nfa, EpsNfa or RedNfa.
     */
    public String prefix() {
	if(isEps())
	    return "EpsNfa";
	else if(isReduced())
	    return "RedNfa";
	return "Nfa";
		
    }

    /** destructor method. This method ensures that when this object
     *  gets deleted, the underlying nfauto gets deleted, too */
    protected void finalize() throws Throwable {
	nfauto_destruct( p );
	super.finalize();
    }

    /** clone method.
     */
    public Object clone() {
	return new Nfa( nfauto_clone( p ) );
    }

    // ----------------------------------      methods for #states, #letters

    /** returns size of the alphabet (number of letters)                    
     */
    public int getNletters()               { return nfauto_getNletters( p ); } 

    /** returns number of states                                            
     */
    public int getNstates()                { return nfauto_getNstates( p );  }


    // ---------------------------------    methods for final/nonfinal states

    /** mark a state as final state 
     */ 
    public void setFinal( int q )              { nfauto_setFinal( p, q ); }

    /** mark a state as nonfinal state 
     */
    public void setNonFinal( int q )           { nfauto_setNonFinal( p, q ); }
    
    /** is state q a final state ?
     */
    public boolean isFinalState( int q ) { return nfauto_isFinalState( p, q );}


    // ----------------------------------           methods for initial states

    /** mark a state as an initial state 
     */
    public void setInit( int q )               { nfauto_setInit( p, q ); }
      
    /** is state q an init(state) ?
     */
    public boolean isInitState( int q )  { return nfauto_isInitState( p, q );}

    /** mark a state as a non initial state
     */                                                          // only Nfa
    public void setNonInit( int q )            { nfauto_setNonInit( p, q ); }


    // -------------------------------------          methods for transitions

    /** make transition in D from F to T with Letter L */
    public void connect(int F,int L,int T)   {nfauto_connect( p, F, L, T );}
    
    /** destroy connection in D from F to T with Letter L */
    public void disconnect(int F,int L,int T){nfauto_disconnect( p, F, L, T );}
    
    /** testcon: TRUE iff connection exists */
    public boolean testcon(int F,int L,int T){return nfauto_testcon(p,F,L,T );}

    /** returns Vector<Vector<int, bool, bool>> containing a transition
     *  from F via L, with <statenumber, init flag, final flag> for
     *  each target.
     *  convenience method. not present on C-level.
     */
    public Vector getcon( int F, int L ) {
	Vector v = new Vector();
	for(int i=0; i<getNstates(); i++) {
	    if( testcon( F, L, i ) ) {
		Vector vv = new Vector(3);
		vv.add(new Integer( i ));
		vv.add(new Boolean( isInitState( i )));
		vv.add(new Boolean( isFinalState( i )));
		v.add( vv );
	    }
	}
	return v;
    }

    /** 
     */
    public boolean isEps() {
	return nfauto_isEps( p );
    }  
    /** 
     */
    public boolean isReduced() {
	return nfauto_isRed( p );
    }    
    /** 
     */
    public void setEps( boolean iseps ) {
	nfauto_setEps( p, iseps );
    }  
    /** 
     */
    public void setReduced( boolean isred ) {
	nfauto_setRed( p, isred );
    }
    

    /** visitor method
     */
    public Object apply( LangDefObjVisitor v ) {
        return v.caseNfa( this );
    }

}
