/* RegExp. java
 *
 * Dez. 2000 Burak Emir
 */

package net.sourceforge.amore.Ldo;

import net.sourceforge.amore.Exceptions.RegExpParseException;

/** this class holds regular expressions. open: unicode ??
 */

public class RegExp extends LangDefObj  {

    private static native int     rex_construct(String s, int sno);
    private static native void    rex_destruct(int p);
    private static native int     rex_getNletters(int p);
    private static native void    rex_setNletters(int p, int i);
    private static native boolean rex_isGen(int p);
    private static native String  rex_getRex(int p);
    private static native String  rex_getExpRex(int p);
    private static native int     rex_clone(int p);

    // private static native boolean

    /** reg. expr. in postfix, abbreviations expanded */
    public String exprex;
    /* abbreviations in infix DISABLED */
    /* public String[] abbr;  */
    /** TRUE iff generalized rex */
    public boolean  generalized;

    // -------  constructors 

    /** construct from c-level RegExp */
    public RegExp( int p ){
	super( p );
    }

    /** throws RegExpParseException if something goes wrong
     */
    public RegExp(String s, int sno /*, int nabbr*/) {
	super( rex_construct(s, sno) );
    }

    /** returns a clone of this object. */
    public Object clone() {
        return new RegExp( rex_clone( p ));
    }

    /** if isGeneralized() returns GenRegExp, otherwise
     *  returns RegExp. used in LangDefObj.getFullName().
     */
    public String prefix() {
	if( isGeneralized() )
	    return "GenRegExp";
	return "RegExp";
    }
    
    /** returns a string containing the regular expression
     */
    public String getRex() {
	return rex_getRex( p );
    }
    
    /** returns a string containing the regular expression,
     *  with all abbreviations expanded.
     */
    public String getExpRex() {
	return rex_getExpRex( p );
    }
    
    /** returns the size of the alphabet
     */
    public int getNletters() {
	return rex_getNletters( p ) ;
    }

    /** sets the size of the alphabet
     */
    public void setNletters( int i ) {
        rex_setNletters( p, i );
    }

    /* 

       // support for abbreviations not implemented yet...      

       public int getNabbr() {
       return abbr.length;
       }
       
       public String getAbbr(int i) {
       return abbr[ i ];
       }
       
       public void setAbbr(int i, String s) {
       abbr[i] = s;
       }
    */

    public boolean isGeneralized() {
	return rex_isGen( p );
    }

    /** visitor method
     */
    public Object apply( LangDefObjVisitor v ) {
        return v.caseRegExp( this );
    }

    
    /** returns a string containing an error message if the parsing
     *  the argument fails, null otherwise 
     *  @param s   the string to parse
     *  @param sno size of the alphabet
     */
    public static String testparse(String s, int sno) {
	try {
	    int i = rex_construct(s, sno);
	} catch (RegExpParseException e) {
	    return e.getMessage();
	}
	return null;
    }
}
