/** \file LibExamples.java
 */

package net.sourceforge.amore ;

import net.sourceforge.amore.InitLibAMoRE ;
import net.sourceforge.amore.Ldo.* ;
import net.sourceforge.amore.Helper.qs ;

import net.sourceforge.amore.Exceptions.RegExpParseException ;

import java.util.Vector ;

/** example code that shows how to define data objects and run algorithms. 
 *
 *  test objects also used for testing AMoREgui.
 *  is a singleton. have a look at the source code
 *  and you will immediately learn all basic functions.
 */

public class LibExamples implements qs {

    static {
	System.out.print("loading AMoRE-java library...");
	System.loadLibrary("AMoRE-java");
	System.out.println("done");
	
	System.out.print("  initializing ...");
        InitLibAMoRE.init();
	System.out.println("ok");
    }

    /** Vector <RegLangSet> (contains some sample RegLangSet objects)*/
    Vector regSet ; 

    /** prints the argument to stdout
     */
    public static void print( LangDefObj o ) {
        DebugPrintVisitor.apply( o );
    }

    /** main method - instantiates LibExamples and prints all 
     *  test objects.
     */
    public static void main(String[] args) {
        LibExamples libEx = getInstance();

        print( d1 );

	System.out.println("cloning ...");
	Dfa tmpd = (Dfa) d1.clone();
	System.out.println("result ...");
	print( tmpd );

        print( d2 );

        print( n1 );

	System.out.println("cloning ...");
	Nfa tmpn = (Nfa) n1.clone();
	System.out.println("result ...");
	print( tmpn );
	
        print( n2 );

        print( rex1 );
	
        print( d3 );
        
        print( n3 );

	String aStr = "abcdefg", bStr="aUbUc[;'{", cStr = "Uab";
	RegExp a,b,c;

	System.out.println("What follows is a demonstration of error"+
                           "\n messages (alphabet a,b) passed from " + 
                           "C through exceptions");
	try {
            System.out.println("Trying to parse \""+aStr+"\"");
            a = new RegExp(aStr,2);    // only a, b allowed
	} catch (RegExpParseException e) {
            System.out.println("error: "+e.getMessage());
	}
	try {
            System.out.println("Trying to parse \""+bStr+"\"");
            b = new RegExp(bStr, 2); // illegal char
        } catch (RegExpParseException e) {
	    System.out.println("error: "+e.getMessage());
	}
	try {
            System.out.println("Trying to parse \""+cStr+"\"");
	    c = new RegExp(cStr, 2);     //illegal syntax
	} catch (RegExpParseException e) {
	    System.out.println("error: "+e.getMessage());
	}

        System.out.println("\ndone.");
    }

    protected static LibExamples instance; /**< singleton instance */

    /** singleton method to prevent multiple instantiations */
    public static LibExamples getInstance() {
	if( instance == null )
	    instance = new LibExamples();
	return instance;
    }

    public static Dfa d1;
    static Dfa d2;
    static Dfa d3;
    static Nfa n1;
    static Nfa n2;  
    static Nfa n3;
    static Monoid mon;

    static Dfa d4;

    /*
    static MutableNfa ne1;
    static MutableNfa ne2;
    */
    static RegExp rex1;
    static RegExp rex2; // gen
    
    /** returns a Vector of several RegLangSet objects
     */
    public Vector getRegLangSetVector() {
	return regSet;
    }
    
    /** initialises objects
     */

    protected LibExamples() {
        
	System.out.print("init'ing LibExamples() ...");

        // this will contain all languages
	regSet = new Vector();
        
        // Dfa "d1" (see below for actual code)
	init_d1();
	regSet.add( new RegLangSet( d1 ,
				    "mydfa1",
				    "this is a very nice and useful dfa"));
        // Dfa "d2" 
	init_d2();
	regSet.add( new RegLangSet( d2 ,
				    "mydfa2",
				    "this is also not bad"));


        // Nfa "n1" 
	init_n1();
        // this RegLangSet contains 3 objects: an nfa...
	RegLangSet r = new RegLangSet( n1 ,
				       "mynfa1",
				       "a nfa plus some other objects");
        // ... its dfa...
        d3 = Transforms.nfa2dfa( n1 ) ;	
	r.addLdo( d3 );
        // ... and its monoid...
	mon = Transforms.dfa2mon( d3 ) ;
	r.addLdo( mon );
	regSet.add( r );

        // Nfa "n2" 
        init_n2();
	regSet.add( new RegLangSet( n2 ,
				    "mynfa2",
				    "another nfa"));

	System.out.println("Hallo vor rex1");

        // RegExp "rex1"
	init_rex1();
	r = new RegLangSet( rex1 ,
                            "myrex1",
                            "this is a rex");

	System.out.println("Hallo vor nfa3");

	// compute nfa
	n3 = Transforms.rex2nfa( rex1 );
	r.addLdo( n3 );
	regSet.add( r );
        

        // RegExp "rex2" (generalised)
	init_rex2();
	r = new RegLangSet( rex2 ,
                            "GENrex",
                            "this is a generalized regexp");
        // compute nfa
	n3 = Transforms.genrex2nfa( rex2 );
	r.addLdo( n3 );
	regSet.add( r );

        r = new RegLangSet( TransformsB.revNfa( n3 ), 
                            "REVmyrex1", 
                            "myrex1 reversed");
        regSet.add( r );
        
	System.out.println(" done");
        /* this would crash because d1 and d2 don't have the same
           alphabets ... */
        /* d4 = BinTransforms.insecDfa( d1, d2 ); */

	/* 
	md4 = MutableDfa ...
	-- TODO: insert demo code for MutableDfa here
	d4 = md4.toDfa()
	*/

	/* 
	mn4 = MutableDfa ...
	-- TODO: insert demo code for MutableNfa here
	n4 = mn4.toNfa()
	*/
    }

    static void init_d1() {
      //System.out.println("LibExamples.init_d1: initializing Dfa d1");
	d1 = new Dfa( 2, 3 );  // 2 letters, 3 states
	d1.setInit( q0 );       // q0 init
	d1.setFinal( q1 );      // q1 final
	d1.connect( q0, a, q1 ); // (q0, a, q1)
	d1.connect( q0, b, q2 ); // (q0, b, q2)
	d1.connect( q1, a, q2 ); // (q1, a, q1)
	d1.connect( q1, b, q2 ); // (q1, b, q2)
	d1.connect( q2, a, q2 ); // (q2, a, q2)
	d1.connect( q2, b, q2 ); // (q2, a, q2)
	d1.setMinimal( false );
    }
    
    static void init_d2() {
      //System.out.println("LibExamples.init_d2: initializing Dfa d2");
	d2 = new Dfa( 1, 2 );  // 1 letter, 2 state
	d2.setInit( q0 );       // q0 init
	d2.setFinal( q1 );      // q1 final
	d2.connect( q0, a, q1 ); // (q0, a, q1)
	d2.connect( q1, a, q1 ); // (q1, a, q1)
	d2.setMinimal( false );
    }

    static void init_n1() {
	n1 = new Nfa( 2, 3 ); // 2 letters, 3 states
	n1.setInit( q0 );
	n1.setFinal( q1 );
	n1.connect( q0, a, q1 );
	n1.connect( q0, a, q2 );
	n1.connect( q0, b, q2 );

	n1.connect( q1, a, q1 );
	n1.connect( q1, b, q2 );
	n1.connect( q2, a, q2 );
	n1.connect( q2, b, q2 );
	n1.setReduced( false );
	n1.setEps( false );
    }

    static void init_n2() {
	n2 = Nfa.constructEpsNfa( 2, 2 ); // 2 letters, 2 states
	n2.setInit( q0 );
	n2.setFinal( q1 );
	n2.connect( q0, a,   q0 );
	n2.connect( q0, eps, q1 );
	n2.connect( q1, a,   q1 );
	n2.connect( q1, b,   q1 );

	// n2.setReduced( false ); FIXME
	// n2.setEps( true );      FIXME
    }

    /*
    static void init_ne1() {
	ne1 = new MutableNfa();
	ne1.addState();
	ne1.addState();
	ne1.removeState();
	// add more weird stuff to test MutableNfa 
    }
    
    static void init_ne2() {
	ne2 = new MutableNfa( n2 );
    }
    */
    static void init_rex1() {
        try {
            rex1 = new RegExp("a*b*", 2);
        } catch (RegExpParseException e) {
            // error handling
	    System.err.println("Error parsing regular expression "+rex1);
            System.err.println(e.getMessage());
            System.exit(0);
        }
    }
    static void init_rex2() {
        try {
            rex2 = new RegExp("ab*-abbb", 2);
        } catch (RegExpParseException e) {
            // error handling
            System.err.println(e.getMessage());
            System.exit(0);
        }
    }
    
}
