/* MutableNfa. java
 *
 * Dez. 2000 Burak Emir
 */

package net.sourceforge.amore.MutableLdo;

import net.sourceforge.amore.Ldo.Nfa ;
import java.util.Vector;
/** intermediary data structure for editing
 */

public class MutableNfa {
    Nfa n;
    public MutableNfa() {
	// default nfa has 1 letter alph, 1 state.
	n = new Nfa( 1, 1 );
    }
    public MutableNfa(int sno) {
	n = new Nfa( sno, 1 );
    }
    public MutableNfa( int nletters, int nstates ) {
	n = new Nfa(  nletters, nstates );
    }
    /** construct with pointer. used to edit an existing nfa. The nfa
     * 	gets cloned
     */
    public MutableNfa(Nfa n) {
      this.n = (Nfa) n.clone();
    }
    
    /** converts this obj to EpsNfa. called when editing is finished
     */
    public Nfa toNfa() {
      return (Nfa) n.clone();
    }

    public boolean isValid() {
	return false;
    }

    // decorator methods

    public void setFinal( int q )               { n.setFinal(q);    } 
    public void setNonFinal( int q )            { n.setNonFinal(q); }
    public void setInit( int q )                { n.setInit(q);    }
    public void setNonInit( int q )             { n.setNonInit(q); }
    public boolean isFinal( int q )        { return n.isFinalState( q );}
    public boolean isInit( int q )         { return n.isInitState( q );}
    public void connect( int F, int L, int T )  {n.connect(F,L,T);   }
    public void disconnect( int F, int L, int T){n.disconnect(F,L,T);}
    public boolean testcon( int F, int L, int T){ return n.testcon(F,L,T);   }
    public int getNletters()         { return n.getNletters();}
    public int getNstates()          { return n.getNstates();}  
    public boolean isEps()              { return n.isEps(); }
    public Vector getcon( int F, int L ) {  return n.getcon( F, L ); }

    /** copies info in this obj to (bigger) obj dest
     */
    private void copyThisTo(Nfa dest) {
      int s = n.getNletters();
      int q = n.getNstates();
      for(int i = 0; i<q; i++) {
	if( n.isFinalState(i) )
	  dest.setFinal( i );
	if( n.isInitState(i) )
	  dest.setInit( i );
	for(int j = isEps()?0:1 ; j<= s; j++) /* letter */
	  for(int k = 0; k<q; k++)
	    if(n.testcon(i,j,k))
	      dest.connect(i,j,k);
	}
    }

    /** adds a new letter at the end.
     */
    public void addLetter() {
	int s = n.getNletters();
	int q = n.getNstates();
	Nfa m = new Nfa( s+1, q );
	copyThisTo( m );
	this.n = m;
    }
    
    /** removes last letter
     */
    public void removeLetter() {
      Nfa m;
	int s = n.getNletters();
	if (s > 1) {
	  int q = n.getNstates();
	  m = new Nfa( s - 1, q );
	  for(int i = 0; i<q; i++) {
	    if( n.isFinalState(i) )
	      m.setFinal( i );
	    if( n.isInitState(i) )
	      m.setInit( i );
	    for(int j = isEps()?0:1; j<= s-1; j++) /* letter */
	      for(int k = 0; k<q; k++)
		if(n.testcon(i,j,k))
		  m.connect(i,j,k);
	  } /* for..*/	  
	this.n = m;
	} /* if.. */
    }
    
    /** adds a new state at the end.
     */
    public void addState() {
	int s = n.getNletters();
	int q = n.getNstates();
	Nfa m = new Nfa( s, q + 1);
	copyThisTo( m );
	this.n = m;
    }
    
    
    /** removes last state
     */
    public void removeState() {
      Nfa m;
	int q = n.getNstates();
	if (q > 1) {
	  int s = n.getNletters();
	  m = new Nfa( s, q-1 );
	  for(int i = 0; i<q-1; i++) {
	    if( n.isFinalState(i) )
	      m.setFinal( i );
	    if( n.isInitState(i) )
	      m.setInit( i );
	    for(int j = isEps()?0:1; j<=s; j++) /* letter */
	      for(int k = 0; k<q-1; k++)
		if(n.testcon(i,j,k))
		  m.connect(i,j,k);
	  } /* for..*/
	this.n = m;
	} /* if..*/
    }

    public String toString() {
      return("<-- MutableNfa containing "+n.toString()+"-->");
    }

}
