/* RegLangSet.java
 *
 * Dez. 2000 Burak Emir
 */

package net.sourceforge.amore;

import net.sourceforge.amore.Ldo.* ;
import java.util.Vector;
import java.util.Iterator;
import java.util.Hashtable ;

/** holds Ldo's which represent the same regular language.
 *  corresponds to "struct regset" in libAMoRE-C.
 */

public class RegLangSet implements Comparable {

    private static native int    regset_construct();
    private static native void   regset_destruct(int p);
    private static native int    regset_getNletters(int p);
    private static native void   regset_setNletters(int p, int sno);
    private static native void   regset_setName(int p, String s);
    private static native int    regset_get_maxlength_name();
    private static native String regset_getName(int p);
    private static native void   regset_setDescr(int p, String s);
    private static native int    regset_get_maxlength_descr();
    private static native String regset_getDescr(int p);
    private static native boolean regset_save(int p, String dir);
    private static native int     regset_load(String name, String dir);
    

    private static native void regset_setInputNfa(int p, int q);
    private static native void regset_setInputDfa(int p, int q);
    private static native void regset_setInputRegExp(int p, int q);
    private static native void regset_addNfa(int p, int q);
    private static native void regset_addRedNfa(int p, int q);
    private static native void regset_addDfa(int p, int q);
    private static native void regset_addMinDfa(int p, int q);
    private static native void regset_addRegExp(int p, int q);
    private static native void regset_addMonoid(int p, int q);

    private static native int regset_getInputIs(int p);
    private static native int regset_getInput(int p);
    
    private static native boolean regset_get_brex(int p); 
    private static native boolean regset_get_bdfa(int p); 
    private static native boolean regset_get_bnfa(int p); 
    private static native boolean regset_get_bmon(int p); 
    private static native int regset_get_lrex(int p);
    private static native int regset_get_ldfa(int p);
    private static native int regset_get_lnfa(int p);
    private static native int regset_get_lmon(int p);

    private int p;
    private LangDefObj input;

    // private Hashtable ldoHash;

    private Class inputClass;
    
    protected RegExp rex;
    protected Nfa    nfa;
    protected Nfa    rednfa;
    protected Dfa    dfa;
    protected Dfa    mindfa;
    protected Monoid mon;

    /*
      //protected RedNfa rnfa;
      //protected MinDfa mdfa;
      protected SfExp  sfexp; ???
    */


    public RegExp getRegExp() { return rex;}
    public Nfa    getNfa()    { return nfa;}
    public Nfa    getRedNfa() { return rednfa;}
    public Dfa    getDfa()    { return dfa;}
    public Dfa    getMinDfa() { return mindfa;}
    public Monoid getMonoid() { return mon;}


    
    /** constructor
     */
    public RegLangSet( LangDefObj o, String name, String descr ) {
	this.p = regset_construct();
	// ldoHash = new Hashtable();
	setName( name );
	setDescr( descr );
	setNletters( o.getNletters() );
	setInput( o );
        observers = new Vector();
    }

    /** construct with pointer to C-level obj 
     */
    private RegLangSet( int p ) {
	this.p = p;
        observers = new Vector();
    }

    // observer methods

    protected Vector observers ;
    
    /** after calling this method, the RegLangSetObserver passed as
     *  argument will be notified whenever an ldo is added to this
     *  RegLangSet */
    public void register( RegLangSetObserver o ) {
        observers.add( o );
    }

    /** after calling this method, the RegLangSetObserver passed as
     *  argument will no longer notified when an ldo is added to
     *  this RegLangSet */
    public void unregister( RegLangSetObserver o ) {
        observers.remove( o );
    }

    /** this method is called to notify the registered observers
     */
    protected void notifyObservers( LangDefObj  ldo ) {
        for(Iterator i = observers.iterator(); i.hasNext(); ) {
            ((RegLangSetObserver) i.next()).ldoAdded( ldo );
        }
    }
  
    /** calls the apropriate regset_addXX function for ldo of type XX
     */
    public void addLdo( LangDefObj o ) {
	// sanity check:  input object must exist before adding 
	if( input==null )
	    throw new RuntimeException(" this RegLangSet does not have a input ldo yet ");
	o.apply( addVisitor );
	o.setRegSet( this );
        notifyObservers( o );
    }

    /** constructor
     */
    public RegLangSet() {
	// System.out.println("in constructor RegLangSet()");
	this.p = regset_construct();
    }

    /** construct with an input object
     */
    public RegLangSet(LangDefObj input) {
	this.p = regset_construct();
	setInput( input );
        setNletters( input.getNletters() );
        setName("unnamed");
    }

    /** construct with an input object
     */
    public RegLangSet(String name, String descr, int sno) {
	this.p = regset_construct();
	setName( name );
	setDescr( descr );
	setNletters( sno );
    }

    /** name must have less characters then the result.
     *  @returns maximal allowed length for name
     */
    public static int getMaxLenName() {
        /* this is determined by underlying C implementation */
        return regset_get_maxlength_name();
    }

    /** descr must have less characters then the result.
     *  @returns maximal allowed length for descr
     */
    public static int getMaxLenDescr() {
        /* this is determined by underlying C implementation */
        return regset_get_maxlength_descr();
    }

    /** sanity check
     */
    public boolean isValid() {
      return ( getNletters() != 0 )
        &&( input != null )
        &&( getName().length() > 0)
        &&( getName().length() < getMaxLenName()) &&
        ( getDescr() != null ? 
          (( getDescr().length() > 0 )
           &&( getDescr().length() < getMaxLenDescr()))
          : true);
    }
  
    /** sets the input ldo on java and c-level
     */
    public void setInput( LangDefObj input ) {
        System.err.println("<"+getName()+">setInput ("+input.getClass()+")");
	this.input = input;
	if( input != null ) {
	    inputClass = input.getClass();
	    input.apply( addVisitor );
	    input.setRegSet( this );
	}
        setNletters( input.getNletters() );
    }
    
    /** returns the input ldo of this object
     */
    public LangDefObj getInput() {
        return input;
    }
    
    /** returns the class of the input ldo
     */
    public Class getInputClass() {
	return inputClass;
    }

    /** sets the input class.
     *  beware! setting input class will cause errors if there
     *   is already an input object present!
     */
    public void setInputClass( Class c ) {
	inputClass = c;
    }
    
    /** get the name.
     */
    public String getName() {
	return regset_getName( p );
    }
    /** set the name.
     */
    public void setName(String str) {
      if( str.length() < getMaxLenName())
        regset_setName(p, str);
      else throw new IllegalArgumentException("name too long!, max is"
                                              + getMaxLenName()
                                              );
    }
    /** set the description string.
     */
    public String getDescr() {
	return regset_getDescr( p );
    }
    /** get the description string.
     */
    public void setDescr(String str) {
        if( str.length() < getMaxLenDescr())
          regset_setDescr(p, str);
        else throw new IllegalArgumentException("descr too long!, max is"
                                                + getMaxLenDescr()
                                                );
    }
    /** get the size of the alphabet.
     */
    public int getNletters() {
	return regset_getNletters(p);
    }
    /** set the size of the alphabet.
     */
    public void setNletters(int sno) {
	regset_setNletters(p, sno);
    }

    /** returns string representation (=name)
     */
    public String toString() {
	return getName();
    }
    
    public int getCpointer() {
	return p;
    }

    /*
    public SfExp getSfExp() {
	return sfexp;
    }
    */

    // *******************************  methods for Comparable interface
    
    /** see Comparable (included in Java API)
     */
    public boolean equals(Object o) {
	RegLangSet rs = (RegLangSet) o;
	return getName().equals( rs.getName() );
    }
    
    /** see Comparable (included in Java API)
     * it is possible to compare two RegLangSet objects 
     * use lexicographical ordering on names
     */
    public int compareTo(Object o) {
	String a = getName();
	RegLangSet rs = (RegLangSet) o;
	String b = rs.getName();
	return( String.CASE_INSENSITIVE_ORDER.compare( a , b ));
    }

    /** save a language
     */
    public boolean save(String dir) {
	return regset_save(p, dir);
    }

    /** load a language.
     *  we have to construct a Java wrapper object for every ldo present
     *  at C level
     *  @param name is the name of the language (without path or ".amr")
     *  @param dir  the directory
     */
    public static RegLangSet load(String name, String dir ) {
	int q = regset_load(name, dir);
	if( q != 0) {
	    RegLangSet rs = new RegLangSet( q );
            int i = regset_getInputIs( q );
            String p = "de.rwth_aachen.i7.de.libAMoRE.Ldo."; // package
            switch(i) {
            case 0: //regex in lan->lrex
            case 10://regex in lan->orex
                rs.setInput( new RegExp( regset_getInput(q) ));
                break;
            case 1: //dfa   in lan->ldfa
            case 11://regex in lan->odfa
                rs.setInput( new Dfa( regset_getInput(q) ));
                break;
            case 2: //nfa   in lan->lnfa
            case 12://nfa   in lan->onfa
            case 13://enfa  in lan->oenfa
                rs.setInput( new Nfa( regset_getInput(q) ));
                break;
            default:
                // error
                System.err.println("error in load lan");
                return null;
            }
            // every newly created object belongs to regset rs
            // we may not forget to set this
            if((i!=0) && regset_get_brex(q)) {
              rs.rex = new RegExp( regset_get_lrex(q) );
                rs.rex.setRegSet( rs );
            }
            if((i!=1) && regset_get_bdfa(q)) {
              Dfa d = new Dfa( regset_get_ldfa(q) );
                d.setRegSet( rs );
                if( d.isMinimal() )
                  rs.mindfa = d;
                else
                  rs.dfa = d;
            }
            if((i!=2) && regset_get_bnfa(q)) {
              rs.nfa = new Nfa( regset_get_lnfa(q) );
              rs.nfa.setRegSet( rs );
            }
            if(regset_get_bmon(q)) {
              rs.mon = new Monoid( regset_get_lmon(q) );
              rs.mon.setRegSet( rs );
            }
            return rs;
        }
	else 
	    return null;
    }

    // what follows is lots of code to save a reference in the right cell

    /** visitor class used to add an ldo without knowing its class
     */
    class AddVisitor implements LangDefObjVisitor {
	public Object caseNfa(Nfa n) { 
	    if( n.isReduced() ) 
		addRedNfa( n );
	    else 
		addNfa( n ); 
	    return null;}
	public Object caseDfa(Dfa d) { 
	    // System.out.println("in AddVisitor caseDfa");
	    if( d.isMinimal() )
		addMinDfa( d );
	    else
		addDfa( d ); 
	    // System.out.println("minimal ?" + d.isMinimal());	    
	    return null;
	}
	public Object caseRegExp(RegExp r) { 
	    addRegExp( r );return null;}
	public Object caseMonoid(Monoid m) { 
	    addMonoid( m );return null;}
    };
    
    /** the AddVisitor used by this object. */
    AddVisitor addVisitor = new AddVisitor();

    protected void addRegExp( RegExp r ) {
	rex = r; 
	regset_addRegExp(p, r.getCpointer() );
    }

    protected void addNfa( Nfa n ) {
	nfa = n;
	regset_addNfa( p, n.getCpointer() );
    }
    protected void addRedNfa( Nfa n ) {
	if( n.isReduced() ) {
	    rednfa = n;
	    regset_addRedNfa( p, n.getCpointer() );
	} else throw new IllegalArgumentException("nfa is not reduced");
    }
    protected void addDfa( Dfa d ) {
      // System.out.println("addDfa");
	dfa = d;
	regset_addDfa( p, d.getCpointer() );
    }
    protected void addMinDfa( Dfa d ) {
        //System.out.println("addMinDfa");
	if( d.isMinimal() ) {
	    mindfa = d;
	    regset_addMinDfa( p, d.getCpointer() );
	} else throw new IllegalArgumentException("dfa is not minimal");
    }
    protected void addMonoid( Monoid m ) {
	mon = m;
	regset_addMonoid( p, m.getCpointer() );
    }

}
