/*
 *  Copyright (c) ?    - 2000 Lehrstuhl fuer Informatik VII, RWTH Aachen
 *  Copyright (c) 2000 - 2002 Burak Emir
 *  This file is part of the libAMoRE library.
 *
 *  libAMoRE is  free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with the GNU C Library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA.  
 */

/* libtest.c
 */

#include "libtest.h"
#include <stdio.h>

/* #include "obj.h" */
#include "vars.h"

#include "rexFromString.h"

#include "buffer.h"

#include "string.h"
#include "nfa2dfa.h"
#include "rex2nfa.h"
#include "genrex2nfa.h"
#include "dfamdfa.h"
#include "dfa2mon.h"


#include "debugPrint.h"

//const char *rexstr = "aba*ab";
const char *rexstr = "(a U b U c)* - ((ac*bc*) U (ac*bc*))";
// test suite 1

regex r1;

static int constr_rexFromString(void) {
  printf("\nconstructing RegExp r:");
  r1 = rexFromString(2, strdup(rexstr));
  printf("result: %s\n",r1->exprex);
  return 1;
}

nfa n1;

__attribute__((unused))
static int transf_rex2nfa(regex r) {
  printf("\n  transforming Rex to Nfa");
  n1 = rex2nfa(r);  
  printf("\n  result:");
  debugPrintNfa( n1 );
  return 1;
}

static int transf_genrex2nfa(regex r) {
  printf("\n  transforming GenRex to Nfa");
  n1 = genrex2nfa(r);  
  printf("\n  result:");
  debugPrintNfa( n1 );
  return 1;
}

dfa d1;

static int transf_nfa2dfa(void) {
  printf("\n  transforming to Dfa");
  d1 = nfa2dfa(n1);
  printf("\n  result:");
  debugPrintDfa( d1 );
  return 1;
}

static void transf_dfa2mindfa(void) {
  printf("\n   minimizing dfa");
  d1 = dfamdfa( d1, TRUE );
  printf("\n  result:");
  debugPrintDfa( d1 );
}

monoid m;

static void transf_dfa2mon(void) {
  printf("\n  transforming dfa to monoid:");

  m = dfa2mon( d1 );
  printf("\n  done.\n");
}
// cloning

__attribute__((unused))
static int clone_dfa(dfa d) {
  dfa p;
  printf("Testing clonedfa...\n");
  p = clonedfa( d );
  printf("result:\n");
  debugPrintDfa( p ); 
  return 1;
}

__attribute__((unused))
static int clone_nfa(void) {
    nfa p;
    printf("Testing clonenfa...\n");
    p = clonenfa( n1 );
    printf("result:\n");
    debugPrintNfa( p );
    return 1;
}

__attribute__((unused))
static void regexp_errorMess(void) {
  regex a,b,c;

  a = rexFromString(2, strdup("abcdefg")); // only a, b allowed
  b = rexFromString(2, strdup("aUbUc[;'{")); // illegal char
  c = rexFromString(2, strdup("Uab"));//illegal syntax
}

int main(void) {

  initbuf();
  printf("libtest\n");
  
  constr_rexFromString();
  transf_genrex2nfa( r1 );
  transf_nfa2dfa();
  transf_dfa2mindfa();
  transf_dfa2mon();

  /*
  clone_nfa( n1 );
  clone_dfa( d1 );
  
  regexp_errorMess();
  */
  return 0;
  // todo: freeing
}
