/*
 *  Copyright (c) ?    - 2000 Lehrstuhl fuer Informatik VII, RWTH Aachen
 *  Copyright (c) 2000 - 2002 Burak Emir
 *  This file is part of the libAMoRE library.
 *
 *  libAMoRE is  free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with the GNU C Library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA.  
 */

/* debugPrint.c
 */

#include "print.h"

#include<stdio.h> /* for printf */

#include "nfa.h"

/** ( global ), for transition arrays */
char itoc[28];

static const char s_true[] = "TRUE ";
static const char s_false[] = "FALSE";

#define DP_BUFSIZE 100

static char buf[ DP_BUFSIZE ];

/** takes array containing init state / final state flags,
 *  prints list of init states to buf
 */
static void printInitStates(mrkfin f, int max) {
      register int i;
      char *b = buf;
      
      *b='\0';
      
      for( i=0; i<=max; i++) {
            if(isinit(f[ i ])) {
                  
                  sprintf(b, "q%d ",i);

                  while(*b++) 
                        ;  /* advance to end of string */

                  if ( b - buf > DP_BUFSIZE - 4) {
                        /* buffer is full... break */
                        sprintf(b, "...");
                        i = max + 1; 
                  }
            }
      }
      if( b>buf ) 
            *--b  = '\0'; /* delete last 2 space */
}

static void printFinalStates(mrkfin f, int max) {
    int i;
    char *b = buf;
    
    *b='\0';

    for( i=0; i<=max; i++) {
	if(isfinal(f[ i ])) {
	    sprintf(b, "q%d ",i);
	    while(*b) { b++; }     /* advance to end of string */
	    if ( b - buf > DP_BUFSIZE - 4) {
		sprintf(b, "...");
		i = max + 1; /* break */;
	    }
	}
    }
    if( b>buf ) 
	*--b = '\0'; /* delete last 2 space */
}

static void printConnectedStates(ndelta delta, int max, int from, int letter) {
    int k;
    char *b = buf;

    *b='\0';
    for( k=0; k<=max; k++) {                /* for all possible to states */
	if(testcon(delta,letter,from,k)) {

	    if( b == buf ) {/* first connection? print arrow */
		sprintf(b, " %c -> {",itoc[ letter ]);
		while(*b) {b++; }   /* advance to end of string */
	    }

	    sprintf(b, "q%d ",k);/* print target */
	    /* printf("\n!q%d b-buf:%d!! buf:X%sX\n",k,b-buf,buf); */
	    while(*b) { b++; }         /* advance to end of string */
	    if ( b - buf > DP_BUFSIZE - 5) { /* check range */
		sprintf(buf - DP_BUFSIZE, "...}");
		k = max + 1; /* break */;
		b = buf; /* ensure that no additional closed brace is written*/
	    }
	}
    }
    
    if( b>buf ) 
	sprintf(b-1, "}"); /*close brace */
}


void printNfa( nfa n ) {
    int i,j;
    printf("<nfa at %p >", n);
    printf("\nqno:  %zu",n->qno);
    printf("  sno:     %zu",n->sno);
    printf("  minimal: %s",n->minimal ? s_true : s_false);
    printf("  is_eps:  %s",n->is_eps ? s_true : s_false);
    /* init and final */
    printInitStates(n->infin, n->qno);
    printf("\ninit states: { %s }  ",buf);
    printFinalStates(n->infin, n->qno);
    printf("final states: { %s }\n",buf);
 
    /* display delta (nondet) */
    printf("%s","\ndelta:\n");  
    fflush(stdout);
    for( i = 0; i<=n->qno; i++) {     /* i = from state */
	printf("q%d  ",i);
	fflush(stdout);
	for( j= n->is_eps?0:1; j<=n->sno; j++) {   /* j = letter     */

	    printConnectedStates(n->delta,n->qno,i,j);
	    printf("%s",buf);
	    /*
	    printf("[ %c ->",itoc[j]);
	    fflush(stdout);
	    foer( k=0; k<=n->qno; k++ ) {/ k = to state   /
		 fflush(stdout);
		 if( testcon(n->delta,j,i,k ))
		    printf("q%d ",k);
	    }
	    printf("]");
	    */
	}
	printf("\n");
    }
    printf("\n");
} /* end debugPrintNfa */

void printDfa( dfa d ) {
    int i,j;
    printf("<dfa at %p >", d);
    printf("\nqno:     %zu",d->qno);
    printf("     sno:     %zu",d->sno);
    printf("\nminimal: %s",d->minimal ? s_true : s_false);
    printf("\ninit:    %zu",d->init);
    /* final */
    printFinalStates(d->final, d->qno);
    printf("\nfinal states: { %s }\n",buf);

    printf("\ndelta:\n");

    /* display delta (det) */
    for( i = 0; i<=d->qno; i++) { /* from state */
	printf("q%d  ",i);
	for( j=1; j<=d->sno; j++) { /* letter */
	    printf("[%c ->",itoc[j]);
	    printf("q%zu ]",d->delta[j][i]);
	}
	printf("\n");
    }
} /* end debugPrintDfa */

void printRegExp( regex r ) {
  printf("<regexp at %p >", r);
  printf("generalized ? %s", (r->grex ? s_true : s_false));
  printf("     sno:     %zu",r->sno);
  printf("\nas String (infix) \"%s\"",r->rex);
  printf("\nexpanded (postfix) \"%s\"\n",r->exprex);
}

void printMonoid( monoid m ) {
  int i = 0;
  printf("<monoid at %p >", m);
  printf("qno (nstates):   %3zu  sno (nletters):    %3zu ",m->qno, m->sno);
  printf("mno (nelements): %3zu  gno (ngenereators) %3zu ",m->mno, m->gno);
  printf("elements {");
  for(i=0;i<m->mno;i++) {
    int j;
    int c = m->lastletter[i];

    if(0 == m->no2length[i])
      printf("%c", itoc[0]);

    for(j=m->no2length[i];j>0;c=m->gensucc[i][0],j--) {
      printf("%c", itoc[c]);
     }
    if(i != m->mno -1)
      printf(",");
  }
  printf("}\n");
}
