/** \file
 *  Copyright (c) 2008        Aidan Delaney <a.j.delaney@brighton.ac.uk>
 *  This file is part of the libAMoRE library test suite.
 *
 *  libAMoRE and associated test suite is  free software; you 
 *  can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with the GNU C Library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA. 
 */
#include "regexp.h"
#include "rexFromString.h"

#include "CUnit/Basic.h"

#include "testdfa.h"
#include "testmon.h"
#include "testsfx.h"

void test_init(void) {
  initbuf();
}

/*
 * Test that a simple regexp can be constructed from a string.
 */
void test_simple_rex(void) {
  char * simple_rex = "(a)*";
  regex r = rexFromString(1, simple_rex);

  CU_ASSERT(1 == r->sno);
  CU_ASSERT(FALSE == r->grex);
  CU_ASSERT(0 == strcmp("a*", r->exprex));
  CU_ASSERT(-1 == r->useda);

  freerex(r);

  simple_rex = "~(ab*)";
  r = rexFromString(2, simple_rex);

  CU_ASSERT(2 == r->sno);
  CU_ASSERT(TRUE == r->grex);
  CU_ASSERT(0 == strcmp("ab*.~", r->exprex));

  freerex(r);
}

void test_rex_failure(void) {
  char * rex = "ab"; // invalid character in regexp
  regex r = rexFromString(1, rex);

  CU_ASSERT(NULL == r);
  CU_ASSERT(0 == strcmp("", (char*) errorParseResult)); // We get no error message.
}

/* The main() function for setting up and running the tests.
 * Returns a CUE_SUCCESS on successful running, another
 * CUnit error code on failure.
 */
int main()
{
  CU_pSuite pSuite = NULL;

  /* initialize the CUnit test registry */
  if (CUE_SUCCESS != CU_initialize_registry())
    return CU_get_error();

  /* add a suite to the registry */
  pSuite = CU_add_suite("test_suite", NULL, NULL);
  if (NULL == pSuite) {
    CU_cleanup_registry();
    return CU_get_error();
  }

  /* add the tests to the suite */
  if ((NULL == CU_add_test(pSuite, "Test init of library", test_init)) ||
      (NULL == CU_add_test(pSuite, "Test some simple regular expressions", test_simple_rex)) ||
      (NULL == CU_add_test(pSuite, "Ensure incorrect regular expressions faile", test_rex_failure)))
    {
      CU_cleanup_registry();
      return CU_get_error();
    }

  pSuite = CU_add_suite("dfa_test_suite", NULL, NULL);
  CU_ADD_TEST(pSuite, test_dfa_from_rex);
  CU_ADD_TEST(pSuite, test_dfa_from_complex_rex);

  pSuite = CU_add_suite("monoid_test_suite", NULL, NULL);
  CU_ADD_TEST(pSuite, test_syntactic_mon_from_rex);
  CU_ADD_TEST(pSuite, test_syntactic_mon_from_complex_rex);

  pSuite = CU_add_suite("star_free_rex_test_suite", NULL, NULL);
  CU_ADD_TEST(pSuite, test_star_free_rex_from_rex);
  CU_ADD_TEST(pSuite, test_star_free_rex_from_complex_rex);

  /* Run all tests using the CUnit Basic interface */
  CU_basic_set_mode(CU_BRM_VERBOSE);
  CU_basic_run_tests();
  CU_cleanup_registry();
  return CU_get_error();
}
