#include "CUnit/Basic.h"

#include "regexp.h"
#include "rex2nfa.h"
#include "rexFromString.h"
#include "dfa.h"
#include "dfamdfa.h"

void test_dfa_from_rex(void) {
  char * rex = "abUba";
  regex r = rexFromString(2, rex);

  nfa n = rex2nfa(r);
  dfa d = nfa2dfa(n);
  dfa m = dfamdfa(d, TRUE);

  CU_ASSERT(2 == m->sno);
  CU_ASSERT(4 == m->qno);

  // ensure that init goes to different states on 'a' and 'b'
  int qa = m->delta[1][m->init];
  int qb = m->delta[2][m->init];

  // looking for the diamond structure inherent in a mdfa accepting a permutation
  int qab = m->delta[2][qa];
  int qba = m->delta[1][qb];

  CU_ASSERT(qa != qb);
  CU_ASSERT(qab == qba);

  freerex(r);
  freenfa(n);
  freedfa(m);
}

void test_dfa_from_complex_rex(void) {
  char * rex = "c*ac*bc*Uc*bc*ac*";
  regex r = rexFromString(3, rex);

  nfa n = rex2nfa(r);
  dfa d = nfa2dfa(n);
  dfa m = dfamdfa(d, TRUE);

  CU_ASSERT(3 == m->sno);
  CU_ASSERT(4 == m->qno);

  // ensure that init goes to different states on 'a' and 'b'
  int qa = m->delta[2][m->init];
  int qb = m->delta[3][m->init];
  int qc = m->delta[1][m->init];

  // looking for the diamond structure inherent in a mdfa accepting a permutation
  int qab = m->delta[3][qa];
  int qba = m->delta[2][qb];

  CU_ASSERT(qa != qb);
  CU_ASSERT(qab == qba);

  freerex(r);
  freenfa(n);
  freedfa(m);
}
